r"""The automatically generated file ``mat24_functions.c`` contains the C
code for the functions exported by the python extension ``mmgroup.mat24``.

These functions are documented in file ``mat24_functions.c``. Here 
just give an overview of the functionality of that module.

File  ``mat24_functions.c`` has been generated from the source file 
``mat24_functions.ske`` using the code generator in module 
``mmgroup.generate_c``. 

The functions in file ``mat24_functions.c`` perform basic computations
in the Golay code, in its cocode, and in the Mathieu group ``Mat24``. 
They also deal with  Parker loop ``Pl`` and with its automorphism group
``AutPl``. A more comfortable python interface to these objects is 
provided by the python classes ``GCode``, ``Cocode``, ``PLoop``, and 
``AutPL`` in module ``mmgroup``.

All C functions in file ``mat24_functions.c`` start with the prefix
``mat24_``. These functions are also called from C functions in other
modules. Therefore we store the binary code for these functions in
a shared library. For some of these functions there are also
macros (defined with ``#define``), starting with ``mat24_def_``.

There is a one-to-one correspondence between the functions in
``mat24_functions.c`` and the function exported from the python
extension ``mmgroup.mat24``, see subsection
*Mapping C functions to python functions* for details.

The python class ``Mat24`` in module ``mmgroup.dev.mat24.mat24_ref``
contains pure python implementations of most functions of the 
``mmgroup.mat24`` extension as class methods. This class is used for 
testing and as a substitute for the ``mmgroup.mat24`` extension in
an early stage of the build process.

In the following subsections the term *C functions* refers to the
C functions in file ``mat24_functions.c``. In the documentation, the 
names of the C functions  are given without the ``mat24_`` prefix.
The term *API reference* means the main document
*The mmgroup API reference* of this project.

The Golay code ``C`` and its cocode ``C*``
.......................................... 

The Mathieu group ``Mat24`` operates as a permutation group on a set 
of 24 elements which we label with numbers 0,...,23 for use in Python 
and C. So it also operates on a vector space ``V = GF(2)**24``, with 
``GF(2) = {0,1}``. Here ``**`` means exponentiation. 

A vector ``v`` in a vector space over ``GF(2)`` is called a bit 
vector. We represent a bit vector as an integer, so that the 
``i``-th  bit of ``v`` (with valence ``2**i``) is the ``i``-th 
component of ``v``. 

The Golay code ``C`` ia a 12-dimensional subspace of ``V`` fixed by
``Mat24``. There are functions for checking and completing codewords 
and for getting the syndrome of a 24-bit vector.

We internally use a basis of ``V`` such that the first 12 basis
vectors are a transversal of the Golay cocode and the last 12 basis 
vectors span the Golay code. These basis vectors are listed in the
API reference.
 
We represent vectors in ``V``, ``C`` and ``C*`` and in the subset of 
octads of ``C`` as follows:

The 759 octads are numbered from 0 to 758. They do not form a vector 
space.
The 2**12 Golay code words are represented as binary numbers 0 to 4095.
The 2**12 cocode words are represented as binary numbers 0 to 4095.
A more detailed description is given in the API reference.

As usual, binary numbers representing bit vectors are added with the 
XOR operation ``^``. Unused high bits in input bit vectors are ignored. 

Functions changing an object from one representation ``xxx`` to another
representation ``yyy`` are named ``xxx_to_yyy``, where ``xxx``, 
``yyy`` is as follows::

  vect:      standard representation of a bit vector in V = GF(2)**24
             coded as a 24-bit integer. 
  vintern:   internal representation a bit vector in V as a vector in
             the basis given above, coded as 24-bit integer. 
  gcode:     representation of a Golay code word in the basis given
             given above, coded as 12-bit integer. 
  octad:     representation as an octad numbered from 0 to 758
             (in lexical order given by representation  'gcode')
  cocode:    representation as a cocode word in the basis given above, 
             coded as 12-bit integer. 


All these representations are given as unsigned integers.
 
We implement the following conversion functions::

    vect_to_vintern, vintern_to_vect, vect_to_cocode, 
    vintern_to_vect, gcode_to_vect, cocode_to_vect.

Here irrelevant bits of the input are ignored. Function
``cocode_to_vect`` returns one of many possible solutions.

In the following functions the input is checked and the function
fails in case of an error::

    vect_to_gcode, vect_to_octad, gcode_to_octad,
    octad_to_vect, octad_to_gcode

In case of failure, these C functions return a special value as 
indicated in the documentation of the function in the .c file.
The corresponding python functions raise ``ValueError`` in case
of failure. 
  
   
Function ``syndrome()`` takes a vector ``v`` and calculates its 
syndrome, which is a vector of minimum weight equivalent to ``v`` 
modulo the Golay code. Function ``cocode_syndrome()`` takes a 
``cocode`` representation of a cocode word instead.

Function ``scalar_prod()`` returns the scalar product of a Golay code
vector in ``gcode`` and a cocode vector in ``cocode`` representation.




The Mathieu group Mat24
....................... 

This class also contains support for the Mathieu group ``Mat24``.
An element of ``Mat24`` can be represented in one of the following ways::

  perm:    Representation as an array of length 24 encoding a 
           permutation of the integers 0,...,23 as a mapping.

  m24num:  Representation as an integer 0 <= i < 244823040. Here i
           is the number of the permutation in lexicographic order.
           So the identity permutation is coded as 0. 

  matrix:  Representation as a 12 x 12 bit matrix acting on the Golay
           code by right multiplication. This matrix acts on a Golay 
           code vectors (given in the 'gcode' representation) by
           right multiplication. 
           Such a matrix is implemented as an array of integers with
           each integer corresponding to a row vector of the matrix. 
           The purpose of this representation is to support 
           the Parker loop and its automorphism group. Therefore a
           row vector is implemented as a 32-bit integer.

We implement the following conversion functions::

    m24num_to_perm, perm_to_m24num, perm_to_matrix, matrix_to_perm.

There is a function ``perm_check()`` for checking if an array of 
length 24 really represents an element of the Mathieu group ``Mat24``.
All other function operating on ``Mat24`` in any way do not check if
their inputs are really in ``Mat24``. They will output garbage on bad
input, but they are not supposed to crash.

The easiest way to create a random element of ``Mat24`` is to create 
a random integer ``0 <= x < 244823040``, and to call function
``m24num_to_perm(x)``. 


Operation of the group ``Mat24`` on vectors
........................................... 

Elements of ``Mat24`` operate from the right on vectors 
``in V = GF(2)**24`` or on Golay code or cocode vectors. 
A function performing such an operation has the name::

    op_<vector>_<group>

where ``<vector>`` indicates the representation of the vector space 
and ``<group>`` indicates the representation of the group. We 
implement the functions::

    op_vect_perm, op_gcode_matrix, op_gcode_perm, op_cocode_perm.

E.g. function ``op_gcode_matrix`` operates on a Golay code word (in 
``gcode`` representation) by right multiplying an element ``m`` 
of ``Mat24`` with it. Here element ``m`` is a  12 times 12 matrix 
(in ``matrix`` representation).  


Group operation in the group ``Mat24``
......................................

Multiplication and inversion in the group ``Mat24`` is supported for 
the permutation representation ``perm``. Therefore we have functions::
 
   mul_perm, inv_perm



The Parker loop ``Pl``
......................

We support the Parker loop ``Pl`` and also its automorphism group
``AutPl``.

An element of ``Pl`` is a pair ``(v, s)``, with ``v`` a Golay code 
word and ``s`` a sign bit, as described in the API reference. We 
represent the element ``(v, s)`` as a  13-bit integer, with ``v``
given by bits 0,...,11 (in ``gcode`` representation) and the sign
``s`` given by bit 12. We call this representation of the Parker 
loop the ``ploop`` representation. So we can convert and element 
of ``C`` in 'gcode' representation to an element of ``Pl`` in 
``ploop`` representation by adjusting the sign in bit 12.

Function ``mul_ploop()`` returns the product of two elements of 
the Parker Loop. Function ``inv_ploop()`` returns the inverse of
ab element of the Parker loop.

Let ``theta`` be the cocycle for the Parker loop defined in the 
API reference. For an element ``v1`` of of ``C`` or ``Pl`` in 
``gcode`` or ``ploop`` representation, the function 
``ploop_theta(v1)`` returns the value ``theta(v1)`` (which is 
in ``C*``) in ``cocode`` representation. Function 
``ploop_cocode(v1, v2)`` returns the value of the coycle 
``theta(v1, v2)``, which is 0 or 1.


The group ``AutPl`` of standard automorphisms of the ``Pl``
...........................................................


An automorphism of the Parker loop is implemented as an array ``a``
of twelve 32-bit integers. The lowest 13 bits of ``a[i]`` encode 
the image of the i-th basis vector of the Parker loop. Here the 
basis of the Parker loop corresponds to the selected basis of the 
Golay code, and each basis vector has positive sign.

The bits ``13,...,24`` of the vectors ``a[i]`` encode a quadratic 
form which facilitates computations in ``AutPl``, as described 
in  section :ref:`implement-autpl-label` in the 
*Guide for developers*.

This representation of ``AutPl`` is called the ``autpl`` 
representation. We only use the ``autpl`` representation for 
elements of ``AutPl``.

Function ``perm_to_autpl(c, p)`` computes an automorphism ``m``
of the Parker loop created from an element ``p`` of ``Mat24``
(given in ``perm`` representation) and a cocode element ``c`` 
(given in ``cocode`` representation). If ``m`` is equal to
the result of ``perm_to_autpl(c, p)``, then we can get back 
``p`` and ``c`` be computing ``p = autpl_to_perm(m)`` and 
``c = autpl_to_cocode(m)``. 

Function ``cocode_to_autpl(c)`` is equivalent to function
``perm_to_autpl(c, p0)``,  where ``p0`` is the identity 
permutation. Note that::

   perm_to_autpl(c, p) = cocode_to_autpl(c) * perm_to_autpl(0, p).   

Here ``perm_to_autpl(0, p)`` is equivalent to 
*standard representative* of ``p`` in ``AutPl``, and 
``cocode_to_autpl(c)`` is a *diagonal automorphism*, as described
in section *Automorphisms of the Parker loop* of the 
API reference.

Function ``op_ploop_autpl(v, m)`` applies Parker loop automorphism 
``m`` to  element ``v`` of ``Pl`` and returns the result. 

Function ``mul_autpl(m1, m2)`` computes the product ``m1 * m2`` of 
the Parker loop automorphisms ``m1`` and ``m2``. Function 
``inv_autpl(m1)`` computes the inverse of the Parker loop 
automorphism ``m1``. 


Auxiliary functions
...................

Here is an overview of some auxiliary functions in this class. 
They are described in the corresponding function documentation:::

  bw24              bit weight of the lowest 24 bits of an integer
  lsbit24           min(24, least significant bit pos.) for an integer 
  gcode_weight      weight of a Golay code word in 'gtype' representation
  vect_to_bit_list  given a bit vector in V, it computes the lists of
                    the positions of the 0 bits and of the 1 bits of v.
  extract_b24       extract bits from bit vector using a 24-bit mask
  spread_b24        spread bit vector according to a 24-bit mask


Internal operation
..................

For switching from the standard representation to the internal
representation we use three tables with ``2**8`` entries of ``24`` bit 
length. For switching back from internal to standard representation 
we use three other tables of the same format. There are also tables 
for computing the syndrome of a vector in ``V`` with respect to the 
Golay code. There is yet another table for the cocycle ``theta`` of
the Parker loop.


Abbreviations for functions and parameters in this class
........................................................

The following list of abbreviations used in names of functions
allows to infer the action of most functions in this module::

    Abbreviation  Meaning                                   Data type

    assoc         associator (in Golay code or Parker loop)
    autpl         automorphism of the Parker loop Pl        uint32_t[12]
    bw24          bit weight of the lowest 24 bits of an int
    cap           intersection (of Golay code elements)
    cocode        element of Golay cocode C*                uint32_t
    cocycle       cocycle:  Pl times Pl  ->  {0,1}
    comm          commutator (in Golay code or Pl)
    gcode         element of Golay code C                   uint32_t
    inv           inversion (in Mat24, Pl, or AutPl)
    lsbit24       least significant bit of an integer, 
                  counting bits 0,...,23 only 
    m24num        number of an element of Mat24             uint32_t         
    matrix        element of Mat24 as binary matrix 
                  acting on the Golay code C                uint32_t[12] 
    mul           multiplication (in Mat24, Pl, or AutPl)
    net           Benes network for an element of Mat24     uint32_t[9]
    octad         number of an octad, i.e. a Golay code
                  element of weight 8                       uint32_t
    op            op_<vector>_<operation> means: 
                  apply <operation> to <vector>  
    op_all        apply operation to all vectors
    perm          element of Mat24 as a permutation         uint8_t[24]
    ploop         element of the Parker loop Pl             uint32_t    
    pow           power operator (in Pl)
    scalar_prod   scalar product (of Golay code and cocode)
    suboctad      suboctad, see function suboctad_to_cocode
    syndrome      syndrome (after decoding Golay code)      uint32_t 
    theta         cocycle theta: Pl -> C^* in Parker loop
    to            <x>_to_<y> means: return representation <y>
                  of an object given in representation <x>
    vect          vector in V = GF(2)**24                   uint32_t
    vintern       vector in V, in internal representation   uint32_t


Conventions for parameters in C functions
.........................................

Parameters of functions are either integers or arrays of integers.
Here all integer types are unsigned and of fixed length, such as
``uint8_t``, ``uint16_t`` or ``uint32_t``. 

The type of a parameter is given by a single letter in the name
of the parameter::

    Name  Meaning                                           Type
    a     array specified in documentation of function      unspecified 
    c     Golay cocode element, represented as 'cocode'     uint32_t
    m     permutation in Mat24 or automorphism of Pl
          represented as a bit matrix                       uint32_t[12]
    p     permutation in Mat24 represented as 'perm'        uint8_t[24] 
    u_<x> unsigned integer, e.g.                            unspecified
           u_exp:   integer denoting an exponent
           u_m24:   number of a permutation in Mat24
           u_octad: number of octad, 0 < u_octad < 259
           u_width: integer denoting a bit width
    v     vector in V, Golay code C or Parker loop Pl 
          represented as vect, vintern, gcode or ploop      uint32_t  

Integer input parameters have name ``u_<x>``, e.g. ``u_m24, u_exp``.
An integer computed by a function is returned as return value.
Input array parameters have a digit as a suffix, e.g.: ``v1, v2, m1``.
Output array parameters have the suffix ``_out``, e.g.: ``p_out``.
Input/output array parameters  have the suffix ``_io``, e.g.: ``m_io``.
         

Mapping C functions to python functions
.......................................


All C functions in module ``mat24_functions`` are documented.
This documentation is not repeated for the corresponding python
functions in module ``mmgroup.mat24``.

The rules for converting a C function to a python function are
as follows:

 * To obtain the name of the python function, strip off the 
   prefix ``mat24_`` from the name of a C functions.

 * To obtain the tuple of input parameters for the python function,
   take the tuple of parameters of the C function and drop are 
   parameters with suffix ``_out``. 

   In the corresponding python function, an iterable object may 
   be passed where the C function expects a pointer to an integer. 
   The minimum length of that iterable object is either clear from 
   the context or documented in the C function.

 * To obtain the return value of the python function, check if 
   the C function has any parameters with suffix ``_out``. Then 
   the sequence of returned objects is the sequence of these 
   parameters, possibly preceded by the return value if the 
   C function returns an integer value. 

   As usual in python, a sequence of length 1 is returned as a
   single object, and a sequence of length >= 1 is returned as
   a tuple. 

   A parameter with suffix ``_out`` is returned as a list of
   integers.

 * A C function may fail under certain circumstances.

   A failure of a function is indicated in the return value
   of the function. Details are given in the documentation of
   the function. The corresponding python function raises 
   ValueError if the C function fails.

 * The python function drops the return value of the C function
   from the sequence of returned python objects in certain cases.

   If the documentation of the C function contains the phrase
   'Returns 0 in case of success and (anything else) in case
   of failure' then the return value is just a status 
   indicator and hence dropped by the python function.

   If the documentation of the C function contains a phrase
   like 'Returns the length of the list ``xxx_out``' then the 
   python function adjusts the length of that returned list 
   appropriately and drops the return value.
   
 * Parameters with suffix ``_io`` refer to pointers in the C
   function and hence to iterables in the corresponding 
   python function. Here the sufix ``_io`` means that
   the function may modify that iterable object.   


"""



# cython: language_level=3

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from libc.stdint cimport uint64_t, uint32_t, uint16_t, uint8_t



cdef extern from "mat24_functions.h":
    const uint32_t MAT24_BASIS[24]


import numpy


"""yet to be documented!!!! """


cimport cython
cimport generators as g



include "../pxd_files/generators.pxi"


def make_table(uint32_t u_box, uint32_t u_exp):
    assert 1 <= u_box <= 5 and 1 <= u_exp <= 2
    a = numpy.zeros(32768, dtype = numpy.uint16) 
    cdef uint16_t[::1] pa = a
    cdef uint32_t length
    length = g.gen_xi_make_table(u_box, u_exp, &pa[0])
    return a[:length]


def invert_table(table, uint32_t n_columns, uint32_t len_result):
    result = numpy.zeros(len_result, dtype = numpy.uint16) 
    cdef uint32_t length = len(table)
    assert length & 31 == 0 and len_result & 31 == 0
    cdef uint16_t[::1] pr = result
    cdef uint16_t[::1] pt = table
    g.gen_xi_invert_table(&pt[0], length, n_columns, &pr[0], len_result)
    return result


def split_table(table, modulus):
    cdef uint32_t length = len(table)
    assert length & 31 == 0
    a = numpy.zeros(length >> 5, dtype = numpy.uint32) 
    cdef uint32_t[::1] pa = a
    cdef uint16_t[::1] pt = table
    g.gen_xi_split_table(&pt[0], length, modulus, &pa[0])
    return table, a


#####################################################################################
# The random generator
#####################################################################################

from numbers import Integral
import numpy as np

ctypedef uint64_t* p_uint64_t

try:
    # The following statement requires python version >= 3.8
    from threading import get_native_id
except ImportError:
    # This works for python version >= 3.3
    from threading import get_ident
    from os import getpid
    def get_native_id():
        return (getpid() << 32) +  get_ident()





_seed_dict = {}


cdef _make_std_rand_gen():
    global _seed_dict
    cdef uint64_t thread_id = get_native_id()
    a = np.zeros(4, dtype = np.uint64)
    cdef uint64_t[::1] a1 = a
    g.gen_rng_seed_rnd(&a1[0], thread_id)
    _seed_dict[thread_id] = a
    return a



@cython.wraparound(False)
@cython.boundscheck(False)
cdef p_uint64_t c_rng_get_seed(seed):
    cdef uint64_t[::1] a1
    if seed is None:
        try:
            a1 = _seed_dict[get_native_id()]
            return &a1[0]
        except KeyError:
            a1 = _make_std_rand_gen()
            return &a1[0]
    else:
        a1 = seed
        assert len(seed) >= 4
        return &a1[0]
       

def rand_get_seed(seed = None):
    cdef uint64_t[::1] a1
    if seed is None:
        try:
            return _seed_dict[get_native_id()]
        except KeyError:
            return _make_std_rand_gen()
    else:
        a1 = seed
        assert len(seed) >= 4
        return seed


def rand_make_seed(value = None):
    a = np.zeros(4, dtype = np.uint64)
    rand_set_seed(a, value)
    return a
        

def rand_set_seed(seed = None, value = None):
    cdef uint64_t[::1] a1
    cdef uint64_t thread_id
    cdef uint64_t seed_value
    global _seed_dict
    if seed is None:
        if value is None:
           _seed_dict.clear()
        else:
             err = "Illegal value for seeding the default random generator."
             raise TypeError, err
        _seed_dict.clear()
    else:
        a1 = seed
        assert len(seed) >= 4
        if value is None:
            thread_id = get_native_id()
            g.gen_rng_seed_rnd(&a1[0], thread_id)
        else:
            seed_value = value
            g.gen_rng_seed_no(&a1[0], seed_value)


@cython.wraparound(False)
@cython.boundscheck(False)
cpdef rand_bytes_modp(uint32_t p, uint32_t num_bytes, seed = None):
    cdef uint32_t p1 = p
    cdef uint32_t n1 = num_bytes
    a = np.zeros(n1, dtype = np.uint8)
    cdef uint8_t[::1] a1 = a 
    cdef p_uint64_t p_seed = c_rng_get_seed(seed)
    cdef int32_t status
    status = g.gen_rng_bytes_modp(p1, &a1[0], n1, p_seed)
    if (status):
        err = "Bad modulus for generating random bytes"
        raise ValueError(err)
    return a

    
@cython.wraparound(False)
@cython.boundscheck(False)
cpdef rand_fill_bytes_modp(uint32_t p, array_bytes, seed = None):
    cdef uint32_t p1 = p
    cdef uint32_t n1 = len(array_bytes)
    cdef uint8_t[::1] a1 = array_bytes 
    cdef p_uint64_t p_seed = c_rng_get_seed(seed)
    cdef int32_t status
    status = g.gen_rng_bytes_modp(p1, &a1[0], n1, p_seed)
    if (status):
        err = "Bad modulus for generating random bytes"
        raise ValueError(err)
    
cpdef uint32_t rand_gen_modp(p, seed = None):
    assert 1 <= p < 0x100000000
    cdef uint32_t p1 = p & 0xffffffff
    cdef p_uint64_t p_seed = c_rng_get_seed(seed)
    cdef int32_t result = g.gen_rng_modp(p1, p_seed)
    return result
     


import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join("..", "..", "src")))

from mmgroup import MMGroup
from mmgroup import MMSpace
# Create an instance M of the monster group
M = MMGroup()
# Create representation space of M (modulo 3)
V = MMSpace(3, M)
# Create an element g of M 
g = M(('d', 0x123), ('p', 217821225))
# Create an vector v in V 
v = V(('T', 712, 13), (2, 'X', 0x345, 13))
# Let g operate on v by right multiplication
print(v * g)



from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



import sys
import warnings
from random import sample, randint
import re


from copy import deepcopy
from collections.abc import Sequence

import pytest



from mmgroup.tests.groups.free_group import FreeGroup











 
####################################################################
####################################################################
### Tests
####################################################################
####################################################################
 

@pytest.mark.auto_group
def test_autogroup():
    print("")
    g_bad = FreeGroup("abc", rewrite={"cb":"bc"} )       
    print(1, g_bad("cb"*30)) 
     
    print(2,  g_bad("bac") )

    g1 = FreeGroup("abc", commute = "ab", bad_commute="ab")
    w1 = g1("ab")
    EXP = 20
    print(3, w1**EXP)
    assert w1**EXP ==  g1("a")**EXP *  g1("b")**EXP 
    g2 = FreeGroup("abc", commute="ab")
    w2 = g2("ab")
    print(4, w2**30)
    assert w2**30 ==  g2("a")**30 *  g2("b")**30 
    w_long = w2**100
    assert w_long == 1 * w_long == w_long * 1
    assert w2 ** (-1)  ==  1 / w2
    assert w_long * w_long**(-1) == g2.neutral()

    g = FreeGroup("abc")
    w1 = g("A")
    w2 = g("a")
    print (5, w1, w2, w1*w2)
    assert w1 * w2 == g.neutral()

    g = FreeGroup("abc")
    w1 = g("AbC")
    w2 = g("cBBaac")
    print (6, w1, w2, w1*w2, w1**(-1)*w2, w1**(w2)) 
    assert  w2**(-1)*w1*w2 == w1**(w2)
    assert  w1*w2 == g("ABaac")
    






       



import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join('..', '..', 'src')))

from mmgroup import GCode, Cocode
from mmgroup.structures.gcode import str_vector24


OUTPUT_FILE = "golay_basis_data.inc"

comment = """

.. comment : This file has been created automatically 
.. comment : by executing docs/source/make_tables.py

"""


gcode_basis = """
The basis vectors ``b_0``,..., ``b_11`` of the Golay code
are given in the following table. The table also shows the
cocyles of the basis vectors as explained below::

         Basis vectors of the Golay code             Cocycle theta
         binary: bit 0,...,23            hex         bit 0,...,11
"""


cocode_basis = """

Representatives of the basis vectors ``c_0``,..., ``c_11`` of the 
Golay cocode are given in the following table::

         Basis vectors of the cocode
         binary: bit 0,...,23            hex
"""




def write_gcode_basis():
    with open(OUTPUT_FILE, "wt") as f:
        f.write(comment)
        f.write(gcode_basis)
        b = GCode.basis
        for i in range(12):
            v = str_vector24(b[i])
            n = "b_%d" % i
            h =  "0x%06x" % b[i]
            th = str_vector24(GCode(1 << i).theta().ord)[:14]
            print("  %4s:  %s,  %s;   %s" % (n, v, h, th), 
                    file = f)
        f.write(cocode_basis)
        b = Cocode.basis
        for i in range(12):
            v = str_vector24(b[i])
            n = "c_%d" % i
            h =  "0x%06x" % b[i]
            print("  %4s:  %s,  %s" % (n, v, h), file = f)
        print("", file = f)
        f.close()



if __name__ == "__main__":
    write_gcode_basis()



r"""This module computes an order vector in a representation of the monster.

In :cite:`Seysen22` we define a vector :math:`v_1` in the 
representation  :math:`\rho_{15}` of the monster group 
:math:`\mathbb{M}` that is used for recognizing elements of the 
subgroup :math:`G_{x0}` of :math:`\mathbb{M}`. Vector :math:`v_1`
is also used for computing the order of an element of 
:math:`\mathbb{M}`, so we call  :math:`v_1` an **order vector**
here.

:math:`v_1` is constructed from two vectors :math:`v_{71} \in \rho_3`
and :math:`v_{94} \in \rho_5` via Chinese remaindering. This module
contains functions for computing vectors  :math:`v_{71}` and
:math:`v_{94}` with the properties required in :cite:`Seysen22`.

These computations are the most time-consuming part of the
computation of :math:`v_1`; and we use multiprocessing for 
performing these computations. 

All computations in the monster group are done with instances 
of class ``MM0``, but not ``MM``. The reason for this is that 
class ``MM`` requires the existence of an order vector.
"""

import os
from random import randint
from collections import OrderedDict
from multiprocessing import Pool, TimeoutError, Lock

import numpy as np

import mmgroup
from mmgroup import structures, MM0, MMV
from mmgroup.mm_space import MMSpace, MMVector
from mmgroup.generators import gen_leech3to2_type4
from mmgroup.generators import gen_leech2_reduce_type4
from mmgroup.clifford12 import leech2matrix_add_eqn
from mmgroup.mm import mm_aux_index_sparse_to_leech2
from mmgroup.mm import mm_aux_mmv_extract_sparse
from mmgroup.mm3 import op_eval_A_rank_mod3 as mm_op3_eval_A_rank_mod3 
from mmgroup.mm15 import op_watermark_A as mm_op15_watermark_A



#######################################################################
# Parallel search function
#######################################################################


def _repeat_f(repetitions, f, args):
    """Auxiliary function for function par_search()
    
    This function executes ``f(*args)`` repeatedly (at most
    ``repetitions`` times) until the return value is a result 
    differnt from None.  It returns the pair ``(result, n)``, 
    where ``result`` is the result of the last execution of 
    ``f``, and ``n`` is the number of ececutions of `f``.
    
    """
    for i in range(repetitions):
        result = f(*args)
        if result is not None:
            return i+1, result
    return repetitions, None         


def par_search(trials, f, *args, **kwds):
    """Repeat executing ``f(*args)`` until execution is successful 
    
    Function ``par_search`` executes ``f(*args)`` repeatedly until 
    a result different from ``None`` is returned. Then function 
    ``par_search`` returns that result.
    
    Execution of ``f`` is repeated (at least) ``trials`` times.
    If ``f`` returns ``None`` in all cases then ``ValueError`` is
    raised. 
    
    Execution of function ``f`` runs in parallel. The number of
    parallel processes is equal to ``os.cpu_count()``. One can 
    use the keyword argument ``processes`` to adjust the number 
    of processes.
    
    In each round function ``f`` is executed ``chunksize`` times,
    where ``chunksize`` can be given by a keyword argument. By
    default we make an educated guess for a good chunk size.
    If ``chunksize`` is too small then there might be too much 
    overhead for setting up the parallelization. If ``chunksize``
    is too large then one must always wait for at least
    ``chunksize`` executions of ``f``, even if the probability
    of success is much larger than ``1 / chunksize``.    
    
    Function ``f`` usually calls a random generator for obtaining
    test values. That random generator should return different 
    random  data when running in different processes. Here a 
    random generator like ``random.randint()`` is ok.

    Preferably, the arguments in ``args`` should be simple objects
    like numbers or strings or lists of such simple objects.  You
    must be aware of the fact that each process works on its own
    copy of the input data.

    If the keyword argument ``verbose`` is True then the 
    function displays an indication of its activity.    
    """
    import math
    trials = max(1, math.ceil(trials))
    processes = kwds.get("processes", 0)
    if not processes:
        processes = os.cpu_count()
    else:
        processes = max(1, min(processes, os.cpu_count()))
    chunksize = kwds.get("chunksize", 0)
    if not chunksize:
        chunksize = max(1, math.ceil(math.sqrt(trials)))
    chunksize = max(1, min(chunksize, trials))
    repetitions = math.ceil(chunksize / processes)
    chunksize = repetitions * processes
    n_iter = math.ceil(trials / chunksize)
    n = 0
    verbose = bool(kwds.get("verbose", False))
    arg_list = [(repetitions, f, args)] * processes
    for i in range(n_iter):        
        with Pool(processes = processes) as pool:
            results = pool.starmap(_repeat_f, arg_list, chunksize = 1)
        pool.join()
        for i, result in results:
            if result is not None:
                n += processes * i
                s = "\rSuitable candidate found after about %d tests"
                if verbose:
                    print(s % n)
                return result
        n += chunksize
        if verbose:
            print("\r%5d candidates tested   " % n, end = "")
    if verbose:
        print("\n")
    err = "No suitable candidate found"
    raise ValueError(err) 


#######################################################################
# Find an element of the monster of a given order
#######################################################################

MMV3 = MMV(3)
MMV15 = MMV(15)


def get_factors(n):
    """Return sorted list of factors of n, for n <= 120"""
    assert 1 <= n < 121, "Number too large"
    return [i for i in range(1, n+1) if n % i == 0]


def rand_mm_element(size = 1):
    """Return a random element ``g`` of the monster group

    Here ``size`` is a measure for the requested word length. 
    of ``g``. The returned random element is not uniform
    distrbuted. It is just sufficiently random, so that
    elements of a high order occur sufficiently often

    ``g`` is returend as an instance of class ``MM0``.
    """    
    data = []
    for i in range(size):
        data += [('t','n'), ('l','n')]
        data += [(s,'r') for s in "yxdp"]
    return MM0(data)

    
def rand_elem_of_order(factors, elem_size):   
    r"""Return random element of the monster group of a given order

    The function creates a random element ``g`` of the monster by 
    calling ``g = rand_mm_element(elem_size)``. It returns ``g`` as 
    a string if ``g`` has order ``o``, where ``o`` is  a multiple of 
    ``factors[-1]`` and ``None`` otherwise. Here ``factors`` must be
    the list of all factors of the value ``factors[-1]`` in natural
    order.

    The order is checked by applying a random vector ``v`` in the 
    representation :math:`\rho_3` and by multplying ``v`` with 
    powers of ``g``. 
    """    
    g = rand_mm_element(elem_size)
    v = MMV3('R')
    w = v.copy()
    for i in range(1, len(factors)-1):
        w.mul_exp(g, factors[i] - factors[i-1])
        if v == w:
            return None
    w.mul_exp(g,  factors[-1] - factors[-2])
    return str(g) if v == w else None
    
    

def find_element_of_order(order, minsize = 1, verbose = 0):
    """Return an element ``g`` the monster of order ``order``.

    We ``g`` that check acts as an element of given ``order`` on a
    vector ``v`` in characteristic ``p``.  So ``g`` has that order
    with a very high probalibility, if ``v`` is selected at random.
    But, theoretically, the order of ``g`` may be a multiple of 
    ``order``. ``v`` should be specified as a list of tuples or set 
    to set to zero for generating a random vector.

    ``g`` is returned as an element of the standard instance of
    class  ``mmroup.MMGroup``.

    Parameter ``minsize`` specifies a measure for requested word 
    length of ``g``.  If no element ``g`` of the requested size
    is found, we look for larger candidates. There is no need 
    for changing the default value of ``minsize``, except if you
    want to make experiments for accelerating special cases.
    """
    if verbose:
        print("Find element g%s of order %s in monster group" 
            % (order, order))
    factors = [0] +  get_factors(order) # [order // x for x in rfactors] + [order]
    if verbose:
        print("Factors =" , factors[1:])
    maxsize = minsize + 10
    for elem_size in range(minsize, maxsize):
        try:
            g = par_search(1000, rand_elem_of_order, factors, 
                   elem_size, chunksize = max(70, order),
                   verbose = verbose
                )
            if verbose:
                 print("Found g%d =" % order, g)
            return g
        except ValueError:
            continue
    err = "No element of order %s found in the monster group"
    raise ValueError(err % order)
 

#######################################################################
# Find a vector stabilizing by an element of order n
#######################################################################

def stabilizer_vector(v, g, n):
    """Compute a vector stabilized by an element of the monster

    Le ``g`` be an element of the monster group of order ``n`` and 
    ``v`` a vector in a represention of the monster. We return the
    vector ``sum(v * g**i for i  in range(n))`` which is stabilized
    by ``g``. We always return ``None`` if that sum is 0 or a 
    multiple of the 1 element in the representation space. The 
    last condition is checked with a fast crude check only.  
    """
    vg = v.copy()
    w = v.copy()
    for i in range(1, n):
        vg *= g 
        w += vg
    assert v == vg * g
    if (w['B'] == 0).all():
        return None
    return w

#######################################################################
# Obtaining a vector mod 3 stable under an element of order 71
#######################################################################




    
def gA_from_type4(v_type4):
    r"""Comupute a group element reducing a type-4 Leech lattice vector 
    
    Let ``v_type_4`` be a vector of type 4 in the Leech lattice mod 2
    in **Leech lattice encoding**.  The function returns an element
    ``g`` of the subgroup :math:`G_{x0}` of the monster that maps
    `v_type_4`` to the standard frame :math:`\Omega`. The group
    element ``g`` is returned as a string.   
    """
    g_data = np.zeros(10, dtype = np.uint32)
    len_g = gen_leech2_reduce_type4(v_type4, g_data)
    gA = MM0('a', g_data[:len_g])
    return str(gA)

def make_v71_sample(g71):
    r"""Compute a vector stabilized by a group element of order 71
    
    Let ``g71`` be an element of the monster group of order 71. The
    function generates  a random vector ``v71`` in the representatoon
    of the monster modulo 3 and computes the vector
    ``w = sum(v71 * g71**i for i in range(71))``. ``w`` is
    stabilized by ``g71``. 
    
    The function returns triple containing ``v71`` if ``w`` is 
    non-trivial and ``w`` satisfies an additional property
    described below. Ohterwise the function returns ``None``. The 
    function succeeds with probability about 1/700; so it must be 
    repeated several times.
    
    The part of the vector ``w`` labelled with the tag ``A`` 
    corresponds to a symmetric bilinear form over the Leech lattice
    modulo 3. The function succeeds if that bilinear form correponding
    to ``w`` has a one-dimesional eigenspace with the eigenvalue 
    ``diag`` containing a type-4 eigenvector in the Leech lattice. 
    
    In case of success we return the triple ``(v71, gA, diag)``. Here 
    ``gA`` is an element  of the subgroup :math:`G_{x0}` of the 
    monster group such that the bilinear form on the Leech lattice 
    corresponding  to `w * gA`` has a type-4 eigenvector in the 
    standard frame math:`\Omega`. 

    The values ``v71`` and ``gA`` are returned as strings. ``diag``
    is returned as an integer.        
    """
    r1 = [("s", x, "r") for x in "XTXTX"]
    g71 = MM0(g71)
    v71 = MMV3(r1)
    w71 = stabilizer_vector(v71, g71, 71)
    if not w71:
        return None
    for diag in range(3):
        v3 = mm_op3_eval_A_rank_mod3(w71.data, diag) & 0xffffffffffff
        if v3:
            v_type4 = gen_leech3to2_type4(v3)
            if v_type4:
                return str(v71), gA_from_type4(v_type4), diag               
    return None 


def find_vector_71_mod3(verbose = 0):
    r"""Compute a vector stabilized by a group element of order 71

    The function computes an element ``g71`` of order 71 of the
    monster and a vector ``v71``  in the representation of the 
    monster modulo 3, such that the vector
    ``w = sum(v71 * g71**i for i in range(71))`` is not trivial. 
    Then ``w`` is stabilized by ``g71``. More precisely, the 
    projection of ``w`` onto the 196883-dimensional irreducible
    representation of tthe monster is not trivial.
    
    The vector ``w`` satisfies the following additional property:
    
    The part of a vector ``w`` labelled with the tag ``A`` always
    corresponds to a symmetric bilinear form over the Leech lattice
    modulo 3. The bilinear form correponding to the computed vector
    ``w`` has a one-dimesional eigenspace with the eigenvalue 
    ``diag`` containing a type-4 eigenvector in the Leech lattice.

    Such a vector ``w`` satisfies the requirements for the
    vector :math:`v_{71} \in \rho_3` in :cite:`Seysen22`. 
    
    The function also computes an element ``gA`` of the subgroup 
    :math:`G_{x0}` of the monster  group such that the bilinear 
    form on the Leech lattice  corresponding  to ``w * gA`` has a 
    type-4 eigenvector in the standard frame math:`\Omega`. 
    
    The function returns the tuple ``(g71, v71, gA, diag)``. 
    Here ``diag`` is returned as an integer; the other components  
    are returned as strings.
    """    
    s_g71 = find_element_of_order(71, verbose = 1)
    assert isinstance(s_g71, str)
    g71 = MM0(s_g71)
    if verbose:
        print("Find vector stabilized by g71") 
    s_v71, s_gA, diag = par_search(2.0e5, make_v71_sample, s_g71, 
        verbose = 1, chunksize = 300)
    if verbose:
        print("g71 =", s_g71)
        print("v71 =", s_v71, "# (mod 3)")
        print("gA =", s_gA)
        print("kernel_diagonal =", diag, "# (mod 3)" )
    return s_g71, s_v71, s_gA, diag


#######################################################################
# Obtaining a vector mod 15 stable under an element of order 94
#######################################################################




def make_v94_sample(s_g94):
    """Auxiliary function for function ``find_vector_94_mod5``
 
    Given ab element ``g94`` of order 94 in the monster, the function
    computes a vector ``v94``  in the representatoon of the  monster 
    modulo 15, such that the vector
    ``w = 3 * sum((-1)**i * v94 * g94**i for i in range(94))``
    is not trivial. Then `w`` is stabilized by ``g94**2``, but not 
    by ``g94``.
    
    The function returns the vector ``v94`` in case of  success
    and ``None in case of failure``.    
    """
    g94 = MM0(s_g94)
    r1 = [("n", x, "r") for x in "XTYTZ"]
    v94 = 3 *  MMV15(r1)
    w = stabilizer_vector(v94 - v94 * g94, g94**2, 47)
    return None if w is None else str(v94)



def do_find_vector_v94_mod5(s_g94, verbose = 0):
    """Compute a vector stabilized by a group element of order 47
 
    The function computes and element ``g94`` of order 94 in the
    monster and a vector ``v94``  in the representatoon of the
    monster modulo 15, such that the vector
    ``w = 3 * sum((-1)**i * v94 * g94**i for i in range(94))``
    is not trivial. Then `w`` is stabilized by ``g94**2``, but not 
    by ``g94``.
    
    The function returns the pair ``(g94, v94)`` as a pair of 
    strings.    
    """
    if verbose:
        print("Find vector v94 stabilized by g94**2, but not by g94")
    try:
        s_v94 = par_search(2, make_v94_sample, s_g94,
            chunksize = 1, verbose = verbose)
        if verbose:
            print("v94 =", s_v94, "# (mod 15)")
        return  s_v94
    except ValueError:
        if verbose:
             print("No suitable candidate for v94 found")
        return None




def find_vector_v94_mod5(verbose = 0):
    r"""Compute a vector stabilized by a group element of order 94

    The function computes an element ``g94`` of order 94 of the
    monster,  and a (sparse) vector ``v94``  in the representatoon 
    of the monster modulo 15, such that the vector
    ``w = 3 * sum((-1)**i * v94 * g94**i for i in range(94))``
    is not trivial. Then `w`` is stabilized by ``g94**2``, but not 
    by ``g94``.
    
    Such a vector ``w`` satifies the requirements for the
    vector :math:`v_{94} \in \rho_5` in :cite:`Seysen22`. 

    The function returns the the pair ``(g94, v94)`` in case of 
    success and a pair ``(None, None)`` in case of failure.  In 
    of success, both objects, ``g94`` and ``v94``, are returned
    as strings. The function fails with a very small probability.  
    """
    s_g94 = find_element_of_order(94, verbose = verbose)
    if s_g94 is None:
        return None, None
    s_v94 = do_find_vector_v94_mod5(s_g94, verbose = verbose)
    return s_g94, s_v94






















r"""Generation of C code for the monster representations modulo ``p``

This script generates the C code for computing in the monster 
representations modulo several small odd numbers ``p`` of the form
``p = 2**l - 1``, ``1 < k <= 8``. The list of integers ``p``
for which code will be genrated is given by the variable
``PRIMES`` in file ``config.py``. The generated C modules are used 
in the python extension ``mmgroup.mm{P}``. Here ``{P}`` is to be
replaced by the actaul modulus ``p``, so the extension for ``p = 3``
has the name ``mmgroup.mm3``. 

Therefore several .c, .h, .pyd files are created, and also one .pyx
file is generated for each modulus ``p``. Then the Cython package  
is used for integrating all C functions into a python extension. 
Here we create one extension for each modulus ``p``. The action of 
each extension is controlled by a .pyx file, which also includes 
the generated .pyd files.

Function ``make_c_h_pxd(p)`` in this module generates the .c, .h
and .pxd files, and function  ``generate_pyx(p, ...)`` generates
the .pyx file for modulus ``p``. 
    
Variable INT_BITS in file config.py should be set to 32 or 64 for
32-bit or 64-bit target systems. There might be performace issues
if INT_BITS is not chosen properly. 


Generating .c files
-------------------

We use the C code generation mechanism in class 
``generate_c.TableGenerator``. There .c and .h files are generated
from files with extension .ske. These .ske files al like .c files, 
but augmented with some code generation statements for entering 
(usually rather large) tables and automatically generated code into 
the .c file to be generated. These .ske files may also have statements 
for automatically generating .h files declaring the exported funtions.  
    
Variable C_FILE_SKELETONS is a list of all .ske files. We will 
create  one  .c file from each .ske file in that list. The variable
C_FILE_SKELETONS is a list of strings where the substring
``'{P}'`` occurs in each string. From the string e.g.
``"mm{P}_op_pi"`` we derive the name ``"mm_op_pi.ske"`` of the
.ske file and the names ``"mm3_op_pi.c"``, ``"mm15_op_pi.c"``
of the C files generated for modulus ``3`` and ``5``.

Entering tables and automatically generated code
................................................

We create an instance ``tg`` of class ``TableGenerator`` for 
generating the .c files. For each modulus ``p`` a specific instance 
of class  ``TableGenerator`` is created. A table generator ``tg`` 
takes two dictionaries ``tables`` and ``directives`` as arguments. 
These dictionaries provide user-defined tables and directives for 
the code generator. There are several table-providing classes that
provide dictionaries ``tables`` and ``directives`` for the 
code genertor ``tg``. Variable ``TABLE_CLASSES`` contains 
the list of table-providing classes. For each of these
table-providing classes an instance is created, one instance for 
each modulus ``p``.  

Each table-providing class has methods ``tables()`` and 
``directives()`` that return the corresponding dictionaries.
The union of the dictionaries  ``tables()`` or ``directives()`` 
of all table-providing classes is passed as the constructor
of the instance ``tg`` as argument ``tables()`` or 
``directives()``, respectively.

A simple example for a table-generating class is class 
``Lsbit24Function`` in module ``mmgroups.dev.mat24.mat24aux``.



Generating the .h and the .pxd file
...................................

For each modulus ``p`` we create a single .h file containing the 
prototypes for the functions in all .c files generated for that
modulus. The name of that .h file is given by the variable 
``H_BASICS_NAME``. In case  ``H_FILE_NAME = "mm{P}_op.h"`` the
.h file ``"mm3_op.h"`` will be created for modulus ``p = 3``.

The content of a .h file is also  generated by the corresponding
instance ``tg`` of class  ``TableGenerator`` described above. 

Each .h file has a prefix given by the string ``H_FILE_BEGIN``
and suffix given by the string ``H_FILE_END``. 

For each .h file the Cython package requires a .pxd file for
generating a python extension. The .pxd file contains essentially 
the same information as the .h file. Method ``generate_pxd()``
of the code generator ``tg`` generates the .pxd file. The name
of the .pxd file is given by ``PXD_FILE_NAME``. The generated
.pxd file is prefixed with the content of the string
``PXD_DECLARATIONS``. In case ``PXD_FILE_NAME = "mm_op{P}.pxd"`` 
the .pxd file ``"mm3_op.pxd"`` will be created for  the
modulus ``p = 3``.


Generating the .pyx file
........................

The functions exported by the python extension ``mmgroup.mm{P}`` 
are just simple wrappers for the C functions generated by the code
generator. These wrappers must be coded in the Cython language.
They are simple enough so that they can be generated automatically
by function  ``pxd_to_pyx()`` in module ``mmgroup.generate_c``.
Function ``pxd_to_pyx()`` takes the .pxd files for one modulus 
``p`` (see last section) as input and it creates the .pyx file
for that modulus with name given by ``PXI_FILE_NAME``. in case
``PYX_FILE_NAME = "mm_op{P}.pyx"`` the file ``"mm_op3.pyx"``
for ``p = 3``.


Augmenting the .pyx file
........................

The .pyx file is augmented by some constants for its specific modulus
``p``. The list of constants is given by ``MM_BASICS_CONSTANTS``.
For each modulus ``p`` we create an instance of class ``MM_Basics`` in
module ``mmgroup.dev.mm_basics.mm_basics``. Here we pass the modulus
``p`` to the constructor of that class. Then the instance of class
``MM_Basics`` contains the required constants as attributes, and we
simply define these contants with its appropriate values in the .pyx
file. So contants with the same name may have different values in
extensions corresponding to different  moduli ``p``. 


 
Location of the output files
............................

The location of the generated output files is controlled by certain
variables in module config.py. Each of these variables specifies the
name of a directory.

Files with extension .c, .h go to the directory ``C_DIR``. Files with
extension .pxd, .pxi, .pyx go to the directory ``PXD_DIR``. 

"""


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import sys
import os
import re
import collections
from collections import defaultdict
import warnings
from numbers import Integral



from config import SRC_DIR, DEV_DIR,  C_DIR, PXD_DIR
from config import REAL_SRC_DIR
from config import INT_BITS, PRIMES
SKE_DIR = os.path.join(DEV_DIR, "mm_op")



if  __name__ == "__main__":
    sys.path.append(REAL_SRC_DIR)
    from mmgroup.generate_c import TableGenerator, make_doc, format_item
    from mmgroup.generate_c import pxd_to_pyx, pxd_to_function_list
    sys.path.pop()


VERBOSE = 0

##########################################################################
# Generating .c files
##########################################################################




H_FILE_SKELETONS = [
    "mm{P}_op_defines",
]

C_FILE_SKELETONS = [
    "mm{P}_op_pi",
    "mm{P}_op_misc",
    "mm{P}_op_xy",
    "mm{P}_op_t",
    "mm{P}_op_xi",
    "mm{P}_op_word",
]

# Additions for the list C_FILE_SKELETONS for spcific values p
C_FILE_SPECIFIC_SKELETONS = defaultdict(list)
C_FILE_SPECIFIC_SKELETONS.update( {
   3:   [
         "mm{P}_op_rank_A",
        ],
   15:  [
         "mm{P}_op_rank_A",
         "mm{P}_op_eval_A",
         "mm{P}_op_eval_X",
        ],
} )


def mm_op_p_sources(p):
    return C_FILE_SKELETONS + C_FILE_SPECIFIC_SKELETONS[p]


##########################################################################
# Generating .h and .pxd files files
##########################################################################


H_FILE_NAME = "mm_op{P}.h"


PXD_FILE_NAME = "mm_op{P}.pxd"




### Generate the headers for specific p

H_FILE_BEGIN = """
// %%GEN h
// This header has been created automatically, do not edit!

#ifndef MM_OP%{P}_H
#define MM_OP%{P}_H

#ifdef __cplusplus
extern "C" {
#endif

#include "mm_basics.h"

"""


H_FILE_END  = """
// %%GEN h
#ifdef __cplusplus
}
#endif
#endif  // #ifndef MM_OP%{P}_H

"""


### Declarations for the generated .pxd files


PYX_DECLARATIONS = """
# cython: language_level=3

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from libc.stdint cimport uint32_t, uint16_t, uint8_t, int32_t
from libc.stdint cimport uint64_t
from libc.stdint cimport uint{INT_BITS}_t as uint_mmv_t


"""


##########################################################################
# Entering contants from class MM_Basics 
##########################################################################

# List of constants added to classes for a specific p from MM_Basics(p)
MM_BASICS_CONSTANTS = [
   "MMV_ENTRIES", "INT_BITS", "LOG_INT_BITS",
   "P", "FIELD_BITS", "LOG_FIELD_BITS", 
   "INT_FIELDS", "LOG_INT_FIELDS", "P_BITS", "MMV_INTS", 
]   

##########################################################################
# Entering tables and automatically generated code
##########################################################################

TABLE_CLASSES = None

def table_classes():
    global TABLE_CLASSES
    if not TABLE_CLASSES is None:
        return  TABLE_CLASSES
    from mmgroup.dev.mm_op import mm_op

    if "mockup" in sys.argv[1:]:
        TABLE_CLASSES = [
            mm_op.Mockup_MM_Op,
        ]
    else:
        from mmgroup.dev.mm_basics import mm_aux, mm_tables, mm_basics
        from mmgroup.dev.mm_basics import mm_tables_xi
        from mmgroup.dev.hadamard import hadamard_t 
        from mmgroup.dev.hadamard import hadamard_xi 
        from mmgroup.dev.mm_op import mm_op_pi, mm_op_xy, mm_op_xi
        TABLE_CLASSES = [
            mm_op.MM_Op,
            mm_op_xi.MonomialOp_xi,
            mm_op_pi.Perm24_Benes,
            mm_op_pi.SmallPerm64,
            mm_op_pi.ScalarProd2048,    
            mm_op_xy.Perm64_xy, 
            hadamard_t.HadamardOpT64,
            hadamard_t.HadamardOpT3,
            hadamard_t.HadamardOpT3A,
            hadamard_xi.HadamardOpXi64,
            hadamard_xi.HadamardOpXi16,
        ]  
    return TABLE_CLASSES

    
##########################################################################
# Generating the .pxd file
##########################################################################



PYX_FILE_NAME = "mm_op{P}.pyx"




##########################################################################
# Generating the .c, .h and .pyx file
##########################################################################



def make_c_h_pxd(p):
    """Generate .c files for a specific p from  .ske files 

    The relevant .ske files are listed in variable
    C_FILE_SKELETONS. One corresponding .h and .pxd file
    is also generated. The .c files and the .h file are written 
    to directory given by C_DIR, the .pxd file is written to the 
    current directory.

    Tables and directives for automatically generated C code
    are taken from the classes listed in TABLE_CLASSES.

    Return pair of lists, one of the .c files and one of the.
    .pxd files generated
    """
    tables = {}
    directives = {}
    global generated_tables
    for table_class in table_classes():
        table_instance = table_class(p)
        tables.update(table_instance.tables)
        directives.update(table_instance.directives)
    tg = TableGenerator(tables, directives, verbose = VERBOSE)
    # first generate C files
    c_files = []
    all_ske_files = [H_FILE_BEGIN]
    for name in H_FILE_SKELETONS:
        ske_file = name.format(P = "") + ".ske"
        ske_path = os.path.join(SKE_DIR, ske_file)
        all_ske_files.append(ske_path)
    for name in mm_op_p_sources(p):
        ske_file = name.format(P = "") + ".ske"
        ske_path = os.path.join(SKE_DIR, ske_file)
        c_file = name.format(P = p) + ".c"
        c_path= os.path.join(C_DIR, c_file)
        print("Creating %s from %s" % (c_file, ske_file))
        tg.generate(ske_path, c_path)
        all_ske_files.append(ske_path)
        c_files.append(c_path)
    all_ske_files.append (H_FILE_END)

    # generate .h and .pxd file
    h_file =  H_FILE_NAME.format(P = p)
    h_path =  os.path.join(C_DIR, h_file)
    pxd_file =  PXD_FILE_NAME.format(P = p)
    print("Creating %s from previous .ske files" % h_file)
    tg.generate(all_ske_files, None, h_path)
    pyx_declarations = PYX_DECLARATIONS.format(
        INT_BITS = INT_BITS, P = p
    )
    tg.generate_pxd(
       os.path.join(PXD_DIR, pxd_file), 
       h_file, pyx_declarations, nogil = True
    )
    return c_files,  [ pxd_file ]







##########################################################################
# Generating the .pyx file
##########################################################################




m_translate_pxd = re.compile(r"mm_op(\d+)_(\w+)")



def translate_pxd(name):
    m = m_translate_pxd.match(name)
    if (m):
        p, method_name = m.groups()
        return  "op_"  + method_name 
    else:
        return name



def generate_pyx(p, pxd_files):
    """The main function of this module for generating code

    This function generates the .pyx file with name given by PYX_FILE_NAME 
    and stores it in the current directory. Cython will use the .pyx file 
    to build a wrapper for all generated C functions. 
    """
    from mmgroup.dev.mm_op import mm_op
    def pyx_comment(text, f):
        print("\n" + "#"*70 + "\n### %s\n" % text + "#"*70 + "\n\n",file=f)

    f_pyx = open(os.path.join(PXD_DIR, PYX_FILE_NAME.format(P = p)), "wt")
    declarations = PYX_DECLARATIONS.format(INT_BITS = INT_BITS, P = 0)
    print(declarations, file = f_pyx)

    for pxd_f in pxd_files:
        pyx_comment("Wrappers for C functions from file %s" % pxd_f, f_pyx)
        pyx_code = pxd_to_pyx(
            os.path.join(PXD_DIR, pxd_f), 
            None, translate_pxd, nogil=True
        )
        print(pyx_code, file = f_pyx)

    pyx_comment("Constants", f_pyx)
    basis = mm_op.MM_Op(p)
    for name in MM_BASICS_CONSTANTS:
        value = getattr(basis, name)
        declaration = "%s = %d\n" % (name, value) 
        print(declaration,  file = f_pyx)

    f_pyx.close()




##########################################################################
# Exported functions for generating code
##########################################################################






if __name__ == "__main__":
    for p in PRIMES:
        c_files,  pxd_files = make_c_h_pxd(p)
        generate_pyx(p, pxd_files)



from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import pytest

import types
import sys
import re
import os
from operator import __or__, __xor__, __and__
from functools import reduce
import random


 
from mmgroup.bitfunctions import v2, lmap, lrange

from mmgroup.generate_c  import prepend_blanks
from mmgroup.dev.mat24.mat24aux import generate_golay24_decode


def lsbit24(x):
    return v2(x | 0x1000000)


from mmgroup import mat24
from mmgroup.dev.mat24.mat24_ref import Mat24


#########################################################################
# Test completion of heptads and octads
#########################################################################


def random_dodecad(gc):
    """Return random dodecad in gcode rpresentation

    """
    for i in range(1000):
        v = random.randint(0,0xfff)
        if gc.gcode_weight(v) == 3:
            return v
    raise ValueError("Could not find a docecad")


def random_subdodecad(gc, odd=False):
    """Return test sample for function  mat24_cocode_as_subdodecad

    The function returns a test sample ``(d, c, i)`` for function
    ``mat24_cocode_as_subdodecad``. Here ``d`` is a dodecad (in gcode
    representation, ``c`` is a cocode element (in cocode 
    representation), and ``i`` is 24  (if ``odd`` is False) of
    an integer ``0 <= i < 24`` representing a bit poition ``otherwise``.

    Then calling ``mat24_cocode_as_subdodecad(d, c, i)`` should succeed.
    """
    d = random_dodecad(gc)
    if odd:
        v = gc.gcode_to_vect(d)
        for j in range(1000):
             i = random.randint(0,23)
             if v & (1 << i) == 0:
                  return d, random.randint(0, 0xfff), i
    else:
        for j in range(1000):
             c = random.randint(0,0xfff)
             if gc.scalar_prod(d ^ 0x800, c) == 0:
                 return d, c, 24
    raise ValueError("Could not find a docecad test sample")


   


def get_testdata(gc):
    for i in range(400):
        for odd in (0, 1):
            yield random_subdodecad(gc, odd)

              
        
  



@pytest.mark.mat24
def test_subdodecad(gc = mat24, verbose = 0):
    for n, (d, c, i)  in enumerate(get_testdata(gc)):
        ok = True
        try:
            res = gc.cocode_as_subdodecad(c, d, i)
        except ValueError:
            err = "Function cocode_as_subdodecad has rturned an error"
            ok = False
        if ok:
            c_res = gc.vect_to_cocode(res) 
            if c_res != c:
               err = ("Cocode Error: expected: 0x%03x, obtained:  0x%03x" %
                   (c, c_res))
               ok = False
        if ok:
            cc = res if i >= 24 else res & ~(1 << i)
            d_bits =  gc.gcode_to_vect(d)
            if cc & d_bits != cc:
               err = "Error: Result is not a subset of the dodecad" 
               ok = False
        if verbose or not ok:
            print("Test", n)
            d_bits =  gc.gcode_to_vect(d)
            print("Dodecad = 0x%03x, bits = 0x%06x" % (d, d_bits)) 
            c_bits = gc.cocode_syndrome(c, gc.lsbit24(d_bits)) 
            print("Cocode  = 0x%03x, bits = 0x%06x, weight = %d" 
                % (c, c_bits, gc.bw24(c_bits))) 
            print("Bit pos. =", i)
            if ok:
                print("Result = 0x%06x" % res) 
            if not ok:
                raise ValueError(err)




             





from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import re
from collections.abc import Iterable
from numbers import Integral, Complex
import math
from random import randint

import numpy as np
from mmgroup.clifford12 import QState12 
from mmgroup.clifford12 import qstate12_unit_matrix 
from mmgroup.clifford12 import qstate12_column_monomial_matrix
from mmgroup.clifford12 import qstate12_row_monomial_matrix
from mmgroup.clifford12 import qstate12_pauli_matrix
from mmgroup.clifford12 import error_string
from mmgroup.clifford12 import qstate12_pauli_vector_mul
from mmgroup.clifford12 import qstate12_pauli_vector_exp


FORMAT_REDUCED = True

class QStateMatrix(QState12):
    r"""This class models a quadratic state matrix
       
    Quadratic state matrices are described in the *API reference* in 
    section **Computation in the Clifford group**.
    
      
    :param rows:
    
        * Binary logarithm of the number of rows of the matrix
 
        * or an instance of class ``QStateMatrix``. Then a deep copy
          of that instance is created.
    
    :type rows: 
    
        ``int`` or and instance of class ``QStateMatrix``
    
    :param cols:
    
        Binary logarithm of the number of columns of the matrix
    
    :type cols: 
    
        A nonnegative ``int``, if parameter ``rows`` is an ``int``
    
    :param data:
    
        The data of the matrix as described below
        
    :param mode:
    
        Evaluated according as described below
        
    :raise:
        * TypeError if ``type(data)`` is not as expected.
        * ValueError if ``data`` cannot be converted to an
          instance of class  ``QStateMatrix``.
    
    
    In terms of the theory of quantum computing, ``rows, cols = 0, n`` 
    creates a column vector or a *-ket* ``|v>`` corresponding to a 
    state of of ``n`` qubits, and ``rows, cols = n, 0`` creates a 
    row vector or a *-bra* ``<v|`` corresponding to a linear function
    on a state of ``n`` qubits.

    If ``rows == cols == n`` and the created ``2**n`` times ``2**n``
    matrix is invertible, then the matrix is (a scalar multiple of) 
    an element of the complex Clifford :math:`\mathcal{X}_{12}` of 
    ``n`` qubits described in :cite:`NRS01`.

    If ``rows`` is an instance of this class then a copy of 
    that instance is created.

    If ``rows`` and ``cols`` are integers then ``data``  may be:

      * ``None`` (default). Then the zero matrix is created.

      * A list of integers. Then that list of integers must encode 
        a valid pair ``(A, Q)`` of bit matrices that make up a 
        state, as described in :ref:`clifford-group-label`,
        subsection **Quadratic state matrices**. 
        In this case parameter ``mode`` is evaluated as follows:
              
          * 1: create matrix ``Q`` from lower triangular part
              
          * 2: create matrix ``Q`` from upper triangular part
               
          * Anything else: matrix ``Q`` must be symmetric.

      * An integer ``v``. Then one of the values ``rows`` and 
        ``cols`` must be zero and the unit (row or column) 
        vector with index ``v`` is created. Here a row vector
        is an ``1`` times ``2**cols`` and a column vector is a
        ``2**rows`` times ``1`` matrix.  
          
    As in ``numpy``, matrix multiplication of quadratic state 
    matrices is done with the ``@`` operator and elementwise
    multiplication of such matrices is done with the ``*``
    operator. A quadratic state matrix may also be multiplied 
    by a scalar. Here the scalar must be zero or of the form:

    .. math::
        2^{e/2} \cdot w \, , \quad  e \in \mathbb{Z}, \;
        w \in \mathbb{C}, \, w^8 = 1   \; .

    Divison by such a scalar is legal.
   
    A matrix of type ``QStateMatrix`` may be indexed with square
    brackets as in ``numpy`` in order to obtain entries, rows, 
    columns or submatrices. Then a complex ``numpy`` array (or a 
    complex number) is returned as in ``numpy``. It is  not 
    possible to change the matrix via item assignment. So the 
    easiest way to obtain a complex version of an instance ``qs`` 
    of type ``QStateMatrix`` is to write ``qs[:,:]``.
    
    A row or column index has a natural interpretation as a
    bit vector. In the theory of quantum computation a bit of
    such a bit vector corresponds to a *qubit*. Let ``qs`` be
    a quadratic state matrix of shape ``(m, n)`` and 
    
    .. math::
        x = \sum_{k=0}^{m-1} 2^k x_k, \; y = \sum_{k=0}^{n-1} 2^k y_k, 
        \, x_k, y_k \in \{0, 1 \}.
        
    Then ``qs[x,y]`` is the entry of matrix ``qs`` with row index 
    corresponding to the bit vector :math:`(x_{m-1},\ldots,x_0)` 
    and column index corresponding to the bit vector 
    :math:`(y_{n-1},\ldots,y_0)`.     
    
    Officially, we support matrices with ``rows, cols <= 12``
    only. Methods of this class might work for slightly 
    larger matrices. Any attempt to constuct a too large
    matrix raises ValueError.
    """
    UNDEF_ROW = 255

    #########################################################################
    # Constructors
    
    def __init__(self, rows, cols = None, data = None, mode = 0):
        super(QStateMatrix, self).__init__(rows, cols, data, mode)

    #########################################################################
    # Elementary operations
        
    def conjugate(self):
        """Compute complex conjugate of the matrix
        
        :return: instance of class ``QStateMatrix``.
        """
        m = QStateMatrix(self) 
        return super(QStateMatrix, m).conjugate()   

    conj = conjugate        

    @property 
    def T(self):
        """Return transposed matrix as in numpy"""   
        return self.copy().transpose()

    @property 
    def H(self):
        """Return conjugate transposed matrix as in numpy"""   
        m = self.copy().transpose()
        return super(QStateMatrix, m).conjugate()    
     
    @property 
    def shape(self):
        """Get shape of the complex matrix represented by the state 
        
        The function returns a pair ``(rows, cols)`` meaning
        that the state corresponds to a complex
        ``2**nrows`` times ``2**ncols`` matrix.         
        """
        return super(QStateMatrix, self).shape


    def copy(self):
        """Return a deep copy of the matrix"""
        return self.__class__(self)
            
    #########################################################################
    # Reshaping a state matrix

    def reshape(self, new_shape, copy = True):
        r"""Reshape matrix to given shape
                        
        :param new_shape:
    
            This shape of the reshaped matrix. It must be a pair of
            integers. A pair ``(n0, n1)`` correponds to a complex
            ``2**n0`` times ``2**n1`` matrix.            
    
        :type new_shape: 
        
            tuple of two integers
    
        :param copy:
        
            A deep copy of the reshaped  matrix is returned if   
            ``copy`` is True (default). Otherwise the matrix is
            reshaped in place.
            
        :type copy: ``bool``
    

        ``new_shape[0] + new_shape[1] = self.shape[0] + self.shape[0]`` 
        must hold.

        If one of the values ``new_shape[0]``, ``new_shape[1]``  
        is ``-1`` then the other value is calculated from the sum 
        ``self.shape[0] + self.shape[1]``. 
        """  
        c = self.copy() if copy else self
        super(QStateMatrix, c).reshape(*new_shape)  
        return c

    #########################################################################
    # Permuting the qubits of the state matrix

       
    def rot_bits(self, rot, nrot, start = 0):    
        r"""Rotate qubit indices of the state matrix ``qs`` in place

        If the state matrix ``qs`` has shape ``(0,n)`` or ``(n,0)`` 
        we rotate  the qubits of ``qs`` in place as follows:
        
        For ``n0 <= i < n0 + nrot``  we map qubit ``i`` to qubit  
        ``n0 + (i + rot) % nrot``. E.g. ``nrot = 3, rot = 1, n0 = 0``
        means qubits are mapped as ``0->1, 1->2, 2->0``.

        Here the entries of the state matrix are labelled by bit 
        vectors of length ``n``. Let ``qs(x[0],...,x[n-1])`` be the 
        entry of the state matrix corresponding to the bit vector
        ``(x[n-1],...,x[0])``.
                
        Then the function changes ``qs to qs'`` with
        ``qs'(...,x[start+nrot],y[nrot-1],...,y[0],x[start-1],...)`` =
        ``qs(...,x[start+nrot],z[nrot-1],...,z[0],x[start-1],...)``,
        where ``z[j] = y[j - rot (mod 3)]``.

        If the state matrix ``qs`` has shape ``(n0, n1)`` then we
        label the entries of state matrix by bit vectors of 
        length ``n0 + n1``, with the ``n1`` highest bits corresponding 
        to  the rows of the matrix, and the ``n0`` lowest bits 
        corresponding to the columns of the matrix.
        """
        return super(QStateMatrix, self).rot_bits(rot, nrot, start)   
        
    def xch_bits(self, sh, mask):    
        r"""Exchange qubit arguments the state  ``qs`` in place

        We label the entries of the state matrix ``qs`` by bit vectors
        and define ``qs(x[n-1],...,x[0])``, with  ``n = n0 + n1``, 
        ``(n0, n1) = qs.shape`` as in method ``rot_bits``.
        
        We exchange qubit ``j`` with argument qubit ``j + sh``
        of the state, if bit ``j`` of ``mask`` is set. If bit ``j``
        of ``mask`` is set then bit ``j + sh`` of ``mask`` must not 
        be set. No ``mask`` bit at  position >= ``n - sh``
        may be set.
        
        E.g.  ``qs.qstate12_xch_bits(1, 0x11)`` changes the 
        state matrix ``qs``  to a state matrix ``qs'``  with
        ``qs'(...,x6,x5,x4,x3,x2,x1,x0.)``  =
        ``qs(...,x6,x4,x5,x3,x2,x0,x1)``.
        """
        return super(QStateMatrix, self).xch_bits(sh, mask)   
   
 
    #########################################################################
    # Applying qubit gates

    def gate_not(self,v):
        r"""Apply not gates to a state
        

        Change the state ``qs``  to a state ``qs'`` with
        ``qs'(x) = qs(x (+) v)``, where ``'(+)'`` is the bitwise 
        xor operation.
        The result is reduced if the input is reduced.
        Computing ``qs.gate_not(1 << j)`` corresponds to
        negating qubit ``j``.
        """
        return super(QStateMatrix, self).gate_not(v)

    def gate_ctrl_not(self, vc, v):
        r"""Apply controlled not gates to a state
               
        Change the state ``qs`` to a state ``qs'``
        with ``qs'(x) = qs(x (+) <vc,x> * v)``  where ``'(+)'`` is 
        the bitwise  xor operation, and ``<.,.>`` is the scalar 
        product of bit vectors.  The result is not reduced. The 
        scalar product of the bit vectors ``j`` and ``jc`` must
        be zero. Otherwise the ``ctrl not`` operation is not
        unitary.
        
        Computing ``qs.gate_ctrl_not(1 << jc, 1 << j)``, 
        for ``jc != j``, corresponds to applying a controlled not  
        gate  to qubit ``j``  contolled by qubit ``jc``. 
        This operation is unitary if and only if the scalar
        product of ``j`` and ``jc`` is zero.
        """
        return super(QStateMatrix, self).gate_ctrl_not(vc, v)

    def gate_phi(self, v, phi):
        r"""Apply phase gate to a state
               
        Change the state ``qs``  to a state ``qs'``
        with ``qs'(x) = qs(x) * sqrt(-1)**(phi * <v,x>)``, where
        ``<.,.>`` is the scalar product of bit vectors and ``'**'`` 
        denotes exponentiation.
        The result is reduced if the input is reduced.
        Computing ``qs.gate_phi(1 << j, phi)`` 
        corresponds to applying a  phase ``(phi * pi/2)``  gate 
        to qubit ``j``. 
        """
        return super(QStateMatrix, self).gate_phi(v, phi)

    def gate_ctrl_phi(self, v1, v2):
        r"""Apply controlled phase gates to a state
               
        Change the state ``qs`` to a state ``qs'``
        with ``qs'(x) = qs(x) * (-1)**(<v1,x>*<v2,x>)``, where 
        ``<.,.>``  is the scalar product of bit vectors and ``'**'``
        denotes exponentiation.
        The result is reduced if the input is reduced.
        Computing ``qs.gate_ctrl_phi(1 << j1, 1 << j2)``
        corresponds to applying a  phase ``pi`` gate to 
        qubit ``j2`` controlled by qubit ``j1``. 
        """
        return super(QStateMatrix, self).gate_ctrl_phi(v1, v2)

    def gate_h(self, v):
        r"""Apply Hadamard gates to a state
            
        Apply a Hadamard gate to all qubits ``j`` of the state ``qs``
        (referred by ``self``) with  ``v & (1 << j) == 1``.
        Aplying a Hadamard gate to gate ``j`` changes a state ``qs``
        to a state ``1/sqrt(2) * qs'``, where
        ``qs'(..,x[j+1],x_j,x[j-1],..)`` = 
        ``qs(..,x[j+1],0,x[j-1],..)``
        + ``(-1)**(x_j) * qs(..,x[j+1],1,x[j-1],..)`` .
        The result is not reduced.
        """
        return super(QStateMatrix, self).gate_h(v)


    #########################################################################
    # Extending and restricting a state
       
    def extend_zero(self, j, nqb, copy = True):    
        r"""Insert ``nqb`` zero qubits at position ``j``. 
         
        Let ``qs`` be the state of shape ``(n0+n1)``, and let 
        ``n = n0 + n1`. We change ``qs`` to the following 
        state ``qs'`` depending on ``n + nqb`` qubits:
        
        ``qs'(x[n-1],...,x[j],y[nqb-1],...,y[0],x[j-1]...,x[0])`` 
        is equal to ``qs(x[n-1],...,x[j],x[j-1]...,x[0])`` if
        ``y[0] = ... = y[nqb-1] = 0`` and equal to zero otherwise.
                
        The function returns *ket*, i.e. a column vector
        of shape ``(n + nqb, 0)``. 
        
        If the input is reduced then the result is also reduced.
        
        If parameter ``copy`` is True (default) then a copy
        of the matrix is modified and returned.
        """
        m = QStateMatrix(self) if copy else self
        super(QStateMatrix, m).extend_zero(j, nqb) 
        return m   

    def extend(self, j,nqb, copy = True):    
        r"""Insert ``nqb`` qubits at position ``j``. . 
              
        Let ``qs`` be the state of shape ``(n0+n1)``, and let 
        ``n = n0 + n1`. We change ``qs`` to the following 
        state ``qs'`` depending on ``n + nqb`` qubits:
        
        ``qs'(x[n-1],...,x[j],y[nqb-1],...,y[0],x[j-1]...,x[0])`` 
        = ``qs(x[n-1],...,x[j],x[j-1]...,x[0])`` .
        
        The function returns *ket*, i.e. a column vector
        of shape ``(n + nqb, 0)``. 
        
        If the input is reduced then the result is also reduced.
        
        If parameter ``copy`` is True (default) then a copy
        of the matrix is modified and returned.
        """
        m = QStateMatrix(self) if copy else self
        super(QStateMatrix, m).extend(j, nqb) 
        return m   


    def restrict_zero(self, j, nqb, copy = True):    
        r"""Restrict ``nqb`` qubits starting at postion ``j`` to ``0``.
               
        Let ``qs`` be the state of shape ``(n0+n1)``, and let 
        ``n = n0 + n1`. We change ``qs`` to the following 
        state ``qs'`` depending on ``n`` qubits:
       
        ``qs'(x[n-1],...,x[0]) is equal to qs(x[n-1],...,x[0])`` if 
        ``x[j] = ... = x[j+nqb-1] = 0`` and equal to zero otherwise.
        We do not change the shape of ``qs``.
        
        The output is reduced if the input is reduced.
        
        If parameter ``copy`` is True (default) then a copy
        of the matrix is modified and returned.
        """
        m = QStateMatrix(self) if copy else self
        super(QStateMatrix, m).restrict_zero(j, nqb) 
        return m   

    def restrict(self, j,nqb, copy = True):    
        r"""This is method ``restrict_zero`` with deletion.
        
        Let ``qs`` be the state of shape ``(n0+n1)``, and let 
        ``n = n0 + n1`. We change ``qs`` to the following 
        state ``qs'`` depending on ``n - nqv`` qubits:

        ``qs'(x[n'-1],...,x[0])`` is equal to 
        ``qs(x[n'-1],...,x[j],0,...,0,x[-1],...,x[0])``.
        
        The function returns *ket*, i.e. a column vector
        of shape ``(n - nqb, 0)``. 
        
        If the input is reduced then the result is also reduced.
        
        If parameter ``copy`` is True (default) then a copy
        of the matrix is modified and returned.
        """
        m = QStateMatrix(self) if copy else self
        super(QStateMatrix, m).restrict(j, nqb) 
        return m   

    def sumup(self, j,nqb, copy = True):    
        r"""Sum up ``nqb`` qubits  starting at position ``j``. 
        
        Let ``qs`` be the state of shape ``(n0+n1)``, and let 
        ``n = n0 + n1``. We change ``qs`` to the following 
        state ``qs'`` depending on ``n - nqb`` qubits:
        
        ``qs'(x[n-1],...,x[j+nqb],x[j-1],...,x[0])`` =
        ``sum_{x[j+nqb-1],...,x[j]}  qs1(x[nn1-1],...,x[0])`` .
        
        The function returns *ket*, i.e. a column vector
        of shape ``(n - nqb, 0)``. 
        
        If the input is reduced then the result is also reduced.
        
        If parameter ``copy`` is True (default) then a copy
        of the matrix is modified and returned.
        """
        m = QStateMatrix(self) if copy else self
        super(QStateMatrix, m).sumup(j, nqb) 
        return m   

    #########################################################################
    # Matrix operations
    
    def lb_rank(self):
        r"""Return binary logarithm of the rank of the matrix

        The rank of a nonzero quadratic state matrix is a power of two.
        The function returns -1 if the matrix is zero.
        """
        return super(QStateMatrix, self).lb_rank()
   
    def lb_norm2(self):    
        r"""Return binary logarithm of squared operator norm.
        
        The operator norm is the largest absolute singular value
        of the matrix. For a quadratic state matrix this is a 
        power of :math:`\sqrt{2}` or zero.
        
        The function returns -1 if the matrix is zero.
        """
        m = self.copy().reduce()
        if m.nrows == 0:
            return -1
        return m.factor[0] + m.nrows - 1 - m.lb_rank() 

    def trace(self):
        """Return the trace of a square matrix.

        The trace is returned as an integer, a floating point
        number, or a complex number.
        """
        return super(QStateMatrix, self).trace()


    def inv(self):
        r"""Return inverse matrix
     
        Returns an instance of class ``QStateMatrix``.
        Raise ValueError if matrix is not invertible.
        """
        return self.copy()._mat_inv()

    def power(self, e):
        """Return the ``e``-th power of the matrix
        
        For a matrix ``m`` the power with exponent
        ``e = 2``  is ``m @ m``, and the power with 
        ``e = -1`` is the inverse matrix of ``m``.                
        """
        if isinstance(e, Integral):
            if e > 1:
                mask = 1 << (e.bit_length() - 2)
                acc = self.copy()
                while mask:
                    acc @= acc
                    if (e & mask):
                        acc @= self
                    mask >>= 1
                return acc
            elif e >= 0:
                nqb = self.shape[0]
                if nqb == self.shape[1]:
                    return self.reduce() if e else qs_unit_matrix(nqb)
                self @ self  # raise shape error
            else:
                return self.inv().power(-e)
        elif isinstance(e, QStateMatrix):
            return e.inv() @ self @ e
        else:
            err = "Illegal exponent type for QStateMatrix"
            raise TypeError(err)


    def order(self, max_order):
        """Try to find the order of a matrix 
        
        If the matrix ``m`` has order ``e`` for ``e <= max_order``,
        i.e. ``m.power(e)`` is the unit matrix, then ``e`` is 
        returned. Otherwise ``ValueError`` is raised.
        
        The function might also succeed if ``e`` is slighty larger
        than ``max_order``. It has run time ``O(max_order**0.5)``.        
        """
        def as_tuple(matrix):
             return tuple(matrix.data) + matrix.factor
        nqb = self.shape[0]
        if not self.lb_rank() == nqb == self.shape[1]:
            err = "QStateMatrix object is not invertible"
            raise ValueError(err)       
        if self.lb_norm2():
            err = "QStateMatrix object has infinite order"
            raise ValueError(err)
        unit = qs_unit_matrix(nqb)
        d = {as_tuple(unit):0}
        n = int(math.sqrt(max_order)) + 2
        m = self.copy()
        for i in range(1, n):
            if m == unit:
                return i
            d[as_tuple(m)] = i 
            m @= self
        m = m.inv()
        acc = m.copy()
        for i in range(1, n):
            t = as_tuple(acc)
            if t in d:
                return i * n + d[t]
            acc @= m
        err = "Order of QStateMatrix object not found" 
        raise ValueError(err)

    #########################################################################
    # Dealing with the Pauli subgroup of the Clifford group
     
    def pauli_vector(self):
        r""" Return matrix as a Pauli vector if it is in the Pauli group

        We represent an element of the Pauli group :math:`\mathcal{P}_{n}` 
        as a product of :math:`2n+2` generators. Each generator may have 
        exponent :math:`0` or :math:`1`. The sequence of these exponents are   
        stored as a bit vector as follows:

           * Bit :math:`2n+1` corresponds to multiplication with 
             the scalar :math:`\sqrt{-1}`.

           * Bit :math:`2n` corresponds to  multiplication with
             the scalar :math:`-1`.

           * Bit :math:`n+i` with :math:`0 \leq i < n` corresponds to 
             a not gate applied to qubit :math:`i`.

           * Bit :math:`i` with :math:`0 \leq i < n` corresponds to 
             a phase :math:`\pi` gate applied to qubit :math:`i`.


        Factors are ordered by bit positions, with the most significant 
        bit position occuring first. 
        
        The function returns the bit vector corresponding to this 
        object as an integer.
        It raises ValueError if the matrix iy not in the Pauli group.
        """
        try:
            return super(QStateMatrix,self).pauli_vector()
        except ValueError:
            err = "QStateMatrix object is not in the Pauli group"
            print("\n%s:\n%s\n" % (err, str(self)))
            raise            

    def pauli_conjugate(self, v, arg = True):
        r"""Conjugate Pauli group elements by the matrix

        The method conjugates an element of the Pauli group or a
        list of such elements with the quadratic matrix and returns
        the congugated Pauli group element (or the list of 
        conjugated elements). All Pauli group elements are encoded
        as in method ``pauli_vector``.

        If :math:`M`is the quadratic state matrix given by
        this object then the Pauli group element :math:`v` is 
        mapped to :math:`M v M^{-1}`.

        Matrix :math:`M` must be in a Clifford group. The function
        raises ValueError if this is not the case.

        In case ``arg = False`` the (complex) arguments of the
        returned Pauli group elements are not computed.
        """
        return super(QStateMatrix,self).pauli_conjugate(v, arg)    


    def to_symplectic(self):
        r"""Convert a quadratic state matrix to a symplectic bit matrix

        Here the quadratic state matrix :math:`Q` given by ``self``
        must be an invertible :math:`2^k\times 2^k` quadratic state
        matrix,  i.e. :math:`Q` must be of shape ``(k,k)``. So 
        :math:`Q` is in the Clifford group  :math:`\mathcal{C}_k`  
        of ``k`` qubits, up to a scalar factor.

        The function computes a :math:`2k \times 2k` bit matrix
        :math:`A` with the following property. 

        If :math:`v` is a bit vector representing an 
        element :math:`g_v` of the Pauli group of ``k`` qubits 
        then :math:`v \cdot A` is the vector representing the
        element :math:`Q \cdot g_v \cdot Q^{-1}` of that
        Pauli group, up to a scalar factor. Here elements of the
        Pauli group are given as integers representing bit vectors
        in the same way as in method ``pauli_vector``, ignoring
        the bits ``2k, 2k+1`` corresponding to the scalar factor.

        So this method computes the natural homomorphism from the 
        Clifford group :math:`\mathcal{C}_k` to the symplectic 
        group :math:`\mbox{S}_{2k}(2)`.

        The function returns :math:`A` as a ``numpy`` array of shape
        ``(2*n)``, with the entries of that array corresponding to 
        the rows of :math:`A`.        
        """
        return super(QStateMatrix,self).to_symplectic()    

        
    #########################################################################
    # Obtaining (complex) entries and submatrices
        
        
    def __getitem__(self, item):
        if not isinstance(item, tuple):
            item = (item,)
        while len(item) < 2:
            item = item + (None,)
        a0, f0 = _as_index_array(item[0], self.shape[0]) 
        a1, f1 = _as_index_array(item[1], self.shape[1])
        if f0 & f1:
            return self.complex()
        a0 = a0 << self.shape[1]
        shape_ =  a0.shape + a1.shape 
        a = np.ravel(a0)[:, np.newaxis] + np.ravel(a1)[ np.newaxis, :] 
        if a.dtype != np.uint32:
            a = np.array(a, dtype = np.uint32)
        a = np.ravel(a, order = "C")
        c = self.entries(a)  
        if len(shape_):        
            return c.reshape(shape_)
        return c[0]

    #########################################################################
    # Formatting a quadratic state matrix

    def show(self, reduced = True):
        """Return a state as a string.
        
        If ``reduced`` is True (default) then the reduced representation
        of the state is displayed. Otherwise the state is displayed 
        'as is'.         
         
        The standard conversion of a state to a string, i.e. method
        ``__str__()``, is equivalent to method ``show()``.         
        """
        if reduced:
            self.reduce()
        return _format_state(self, reduced)


    def __str__(self):
        self.reduce()
        return _format_state(self, True)




####################################################################
# Formatting a QStateMatrix object
####################################################################


PHASE = [ ("",""),  (""," * (1+1j)"),  (""," * 1j"), ("-"," * (1-1j)"),
          ("-",""), ("-"," * (1+1j)"), ("-"," * 1j"), (""," * (1-1j)"), 
]

def _format_scalar(e, phase):
    if (phase & 1): e -= 1
    prefix, suffix = PHASE[phase]
    if 0 <= e <= 12 and not e & 1:
        s_exp = str(int(2**(e/2)))
    else:
        s_exp = ("2**%.1f" if e & 1 else "2**%d") % (e/2)
        if prefix:
            s_exp = "(" + s_exp + ")"
    return prefix + s_exp + suffix  
  
BRACKETS = { 
   (0,0):("  <","scalar",">"),   # a scalar
   (0,1):("  <","","|"),         # a row vector or a  *bra-*
   (1,0):("  |","",">"),         # a column vector or a  *-ket*
   (1,1):("  |", "><","|")       # a matrix
}


def binary(n, start, length, leading_zeros = False):
    if length == 0:
        return ""
    n = (n >> start) & ((1 << length) - 1)
    b = format(n, "b")
    c = "0" if leading_zeros else " "
    return c * (length - len(b)) + b


D_PM = { '0':'+', '1':'-'}
D_J = { '+':'.', '-':'j'}


def binary_q(n, start, length, pos, reduced = True):
    def pm_repl(m):
        return "-" if  m.group(0) == "1" else "+"
    def j_repl(m):
        return "j" if  m.group(0) == "1" else "."
    if length < 2:
        return ""
    s =  binary(n, start, length, leading_zeros = True) 
    s = [D_PM[x] for x in s[::-1]]
    s[pos] = D_J[s[pos]] if pos else '.'
    s = "".join(s)
    if reduced:
        return s[:pos+1] + " " * (length - pos - 1)
    return s

                      
def _format_data(data, rows, cols, reduced = False):
    s = ""
    nrows = len(data)
    if len(data) < 2 and  rows + cols == 0:
        return s
    if len(data) == 0:
         data = [0]
    left, mid, right = BRACKETS[bool(rows), bool(cols)]
    left_bl = " " * len(left)
    mid_bl = " " * len(mid)
    right_bl = " " * len(right)
    nq, nd = len(data) - 1, 0
    for i, d in enumerate(data):
        rc = binary(d, 0, rows + cols, not reduced or i == 0)
        r, c = rc[:rows], rc[rows:]
        q = binary_q(d, rows + cols, len(data), i, reduced)
        sq, sd = "  " + q,  left + r + mid + c + right  + "\n"
        if reduced:
            s += sq[:len(sq) - nq] + "  " + sd[nd:]
            nq, nd = nq - 1, nd + 1
        else:
            s += sq +  "  " + sd
        left, mid, right = left_bl, mid_bl, right_bl
    return s
        

STATE_TYPE = { (0,0) : ("QState scalar"), 
               (0,1) : ("QState row vector"), 
               (1,0) : ("QState column vector"),
               (1,1) : ("QState matrix")
}    
        
    
def _format_state(q, reduced = False):
    """Return  a ``QStateMatrix`` object as a string."""
    rows, cols = q.shape
    try:
        data = q.data
    except ValueError:
        print("\nError: Bad instance of class QStateMatrix:\n") 
        print(_format_state_raw(q))        
        raise    
        
    e = q.factor
    str_e = _format_scalar(*e) if len (data) else "0"
    str_data = _format_data(data, rows, cols, reduced)   
    qtype = STATE_TYPE[bool(rows), bool(cols)]
    s = "<%s %s" % (qtype, str_e)
    if len(str_data):
       s += " *\n" + str_data 
    s += ">\n"                  
    return s

def _format_state_raw(q):
    s = ("QStateMatrix, shape = %s, factor = %s, nrows = %s, cols = %s\n"
        % (q.shape, q.factor, q.nrows, q.ncols))
    s += "data:\n"
    length = q.nrows + q.ncols
    for d in q.raw_data[:q.nrows]:
        s += binary(d, 0, length) + "\n"
    return s
        

####################################################################
# Computing an array of indices
####################################################################
        
        
def _as_index_array(data, nqb):
    """Convert an index ``data`` to an array of indices.
    
    ``data`` in an object for indexing a one-dimesnional numpy
    array of length ``nqb``. 
    
    Return a pair ``(a, full)`` where ``a`` is a numpy array 
    containing the indices and ``full`` is True iff
    ``a`` contains the data ``range(1 << nqb)``.
    """
    mask = (1 << nqb) - 1
    if isinstance(data, Integral):
        return np.array(data & mask, dtype = np.uint32), False
    if data is None:
        return np.arange(1 << nqb, dtype = np.uint32), True
    if isinstance(data, slice):
        ind = data.indices(1 << nqb)
        full = ind == (0, 1 << nqb, 1) 
        return np.arange(*ind, dtype = np.uint32), full
    ind = np.array(data, dype = np.uint32, copy = False) 
    if len(ind.shape) > 1:
        err = "Bad index type for QState12 array"
        raise TypeError(err)
    return  ind & mask, False 
    
    
    

    
####################################################################
# Creating specific state matrices of type QStateMatrix
####################################################################



def qs_unit_matrix(nqb):
    """Return unit matrix as an instance of class ``QStateMatrix``
    
    The returned unit matrix has shape ``(nqb, nqb)``. So it
    represents a ``2**nqb`` times ``2**nqb`` unit matrix.
    """
    qs = QStateMatrix(nqb, nqb) 
    qstate12_unit_matrix(qs, nqb)
    return qs
    
def qs_rand_matrix(rows, cols, data_rows):
    limit = (1 << (rows  + cols + data_rows)) - 1 
    data = [randint(0, limit) for i in range(data_rows)]
    return QStateMatrix(rows, cols, data, mode = 1)
          
def qs_column_monomial_matrix(data):
    nqb = len(data) - 1
    qs = QStateMatrix(nqb, nqb) 
    qstate12_column_monomial_matrix(qs, nqb, data)
    return qs
    
def qs_row_monomial_matrix(data):
    nqb = len(data) - 1
    qs = QStateMatrix(nqb, nqb) 
    qstate12_row_monomial_matrix(qs, nqb, data)
    return qs


def qs_pauli_matrix(nqb, v):
    qs = QStateMatrix(nqb, nqb)
    qstate12_pauli_matrix(qs, nqb, v)    
    return qs
    
def qs_ctrl_not_matrix(nqb, vc, v):
    """Return Transformation matrix for ctrl-not gate
    
    ``nqb`` is the rank of the matrix.
    TODO: ``vc, v`` yet to be documented.

    """
    qs = qs_unit_matrix(nqb)
    mask = (1 << nqb) - 1
    qs.gate_ctrl_not(vc & mask, v & mask) 
    return qs.reduce()    

def qs_phi_matrix(nqb, v, phi):
    qs = qs_unit_matrix(nqb)
    qs.gate_phi(v << nqb, phi) 
    return qs    

def qs_ctrl_phi_matrix(nqb, v1, v2):
    qs = qs_unit_matrix(nqb)
    qs.gate_ctrl_phi(v1 << nqb, v2) 
    return qs    

def qs_hadamard_matrix(nqb, v):
    qs = qs_unit_matrix(nqb)
    qs.gate_h(v)    
    return qs.reduce()


####################################################################
# Some wrappers for functions written in C
####################################################################

    
def flat_product(a, b, nqb, nc):
    return a.copy().qstate12_product(b.copy(), nqb, nc)


def pauli_vector_mul(nqb, v1, v2):
    return qstate12_pauli_vector_mul(nqb, v1, v2)

def pauli_vector_exp(nqb, v, e):
    return qstate12_pauli_vector_exp(nqb, v, e)



from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


"""Auxilary functions for group representations"""

import numpy as np 


def sgn(s):
    """return (-1)**(s)"""
    return 1 - ((s & 1) << 1)

def zero_vector(length, *data):
    """Return zero numpy vector of given length and type np.int32"""
    return np.zeros(length, dtype = np.int32)

 

#234567890123456789012345678901234567890123456789012345678901234567890

def pm_mat_from_function(f, l):
    """Create an  l times l matrix from a function f.

    The returned matrix m has entries m[i,j] = (-1)**f(i,j),  i, j 
    = 0,...,l-1. It is a numpy array of shape (l,l) and type np.int32.
    """
    a = np.zeros((l,l), dtype = np.int32)
    for i in range(l): 
        for j in range(l): 
            a[i,j] =  1 - 2 * (f(i,j) & 1)
    return a


def pm_diag_from_function(f, l):
    """Create an  l times l diagonal matrix from a function f.

    The returned matrix m has entries m[i,i] = (-1)**f(i),   
    i = 0,...,l-1, and zero entries for i != j.
    It is a numpy array of shape (l,l) and type np.int32.
    """
    a = np.zeros((l,l), dtype = np.int32)
    for i in range(l): 
        a[i,i] =  1 - 2 * (f(i) & 1)
    return a




# cython: language_level=3

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from libc.stdint cimport uint32_t, uint16_t, uint8_t;




include "../pxd_files/mat24_functions.pxd"


cdef extern from "mat24_functions.h":
    const uint32_t MAT24_BASIS[24]
    const uint32_t MAT24_RECIP_BASIS[32]
    const uint16_t MAT24_SYNDROME_TABLE[0x800]




MAT24_ORDER =  244823040 


###########################################################################
# Some general bit operations
###########################################################################


def lsbit24(uint32_t v1):
    return mat24_lsbit24(v1) 


def bw24(uint32_t v1):
    return mat24_bw24(v1) 


###########################################################################
# Basis of Golay cocode and code
###########################################################################

cdef int _i 
basis =  [MAT24_BASIS[_i] for _i in range(24)]
recip_basis =  [MAT24_RECIP_BASIS[_i] for _i in range(24)]



###########################################################################
# Conversion between bit vectors of GF(2)**24
###########################################################################


def vect_to_bit_list(uint32_t v1):
    cdef uint8_t a[24]
    cdef uint32_t w
    w = mat24_vect_to_bit_list(v1, &a[0])
    return w, list(a[:24])



def vect_to_list(uint32_t v1, uint32_t u_len):
    cdef uint8_t a[24]
    cdef uint32_t w
    w = mat24_vect_to_list(v1, u_len, &a[0])
    return list(a[:w])



def extract_b24(uint32_t v1, uint32_t u_mask):
    return mat24_extract_b24(v1, u_mask)


def spread_b24(uint32_t v1, uint32_t u_mask):
    return mat24_spread_b24(v1, u_mask)



###########################################################################
# Conversion between representations of GF(2)**24, Golay code, etc.
###########################################################################



def vect_to_vintern(uint32_t v1): 
    return mat24_vect_to_vintern(v1)



def vintern_to_vect(uint32_t v1):
    return  mat24_vintern_to_vect(v1)


def vect_to_cocode(uint32_t v1):
    return mat24_vect_to_cocode(v1)


def gcode_to_vect(uint32_t v1):
    return mat24_gcode_to_vect(v1)


def cocode_to_vect(uint32_t c1):
    return mat24_cocode_to_vect(c1)


def vect_to_gcode(uint32_t v1):
    cdef uint32_t cn
    cn = mat24_vect_to_gcode(v1)
    if cn & 0xff000000:
        err = "Bit vector is not a Golay code word"
        raise ValueError, err
    return cn


def gcode_to_octad(uint32_t v1, uint32_t u_strict = 1):
    cdef uint32_t cn
    cn = mat24_gcode_to_octad(v1, u_strict)
    if cn & 0xff000000:
        err = "Golay code word is not an octad"
        raise ValueError, err
    return cn


def vect_to_octad(uint32_t v1, uint32_t u_strict = 1):
    cdef uint32_t v
    v = mat24_vect_to_octad(v1, u_strict)
    if v & 0xff000000:
        err = "Bit vector is not an octad"
        raise ValueError, err
    return v




def octad_to_gcode(uint32_t u_octad):
    cdef uint32_t cn
    cn = mat24_octad_to_gcode(u_octad)
    if cn & 0xff000000:
        err = "Illegal octad number"
        raise ValueError, err
    return cn


def octad_to_vect(uint32_t u_octad):
    cdef uint32_t cn
    cn = mat24_octad_to_vect(u_octad)
    if cn & 0xff000000:
        err = "Illegal octad number"
        raise ValueError, err
    return cn




###########################################################################
# Golay code syndromes and weights
###########################################################################


def syndrome_table(c1):
    return MAT24_SYNDROME_TABLE[c1 & 0x7ff]

def cocode_syndrome(uint32_t c1,  uint32_t u_tetrad = 24):
    cdef uint32_t res
    res = mat24_cocode_syndrome(c1, u_tetrad)
    if res == 0xffffffff:
        raise ValueError("Golay code syndrome is not unique")
    return res


def syndrome(uint32_t v1,  uint32_t u_tetrad = 24):
    cdef uint32_t res
    res =  mat24_syndrome(v1, u_tetrad)
    if res == 0xffffffff:
        raise ValueError("Golay code syndrome is not unique")
    return res



def vect_type(v1):
    cdef uint32_t res
    res = Mat24Sub_vtype(v1)
    return ( res >> 5, (res >> 2) & 7, res  & 3 )




def gcode_weight(uint32_t v1):
    return mat24_gcode_weight(v1)


def gcode_to_bit_list(uint32_t  v1):
    cdef uint8_t data[24]
    cdef uint32_t length
    length = mat24_gcode_to_bit_list(v1, &data[0])
    return list(data[:length])



def cocode_weight(uint32_t  c1):
    return mat24_cocode_weight(c1)


def cocode_to_bit_list(uint32_t  c1,  uint32_t u_tetrad = 24):
    cdef uint8_t data[4]
    cdef uint32_t length
    length = mat24_cocode_to_bit_list(c1, u_tetrad, &data[0])
    if length == 0xffffffff:
        raise ValueError("Golay code syndrome is not unique")
    return list(data[:length])
    
    
def cocode_to_sextet(uint32_t  c1):
    cdef uint8_t data[24]
    cdef uint32_t result
    result = mat24_cocode_to_sextet(c1, &data[0])
    if result == 0xffffffff:
        raise ValueError("Golay cocode word is not a sextet")
    return list(data[:24])


############################################################################
# Scalar product of Golay code and cocode
############################################################################



def scalar_prod(uint32_t v1, uint32_t c1):
    return mat24_scalar_prod(v1, c1)

############################################################################
# Conversion from and to suboctads
############################################################################


def suboctad_to_cocode(uint32_t u_sub, uint32_t v1):
    cdef uint32_t res
    res =  mat24_suboctad_to_cocode(u_sub, v1)
    if res & 0xfffff000:
        raise ValueError("Attempt to compute suboctad of a non-octad")
    return res


def cocode_to_suboctad(uint32_t c1, uint32_t v1):
    cdef uint32_t res
    res =  mat24_cocode_to_suboctad(c1, v1)
    if res & 0xfffff000:
        raise ValueError("Octad/suboctad mismatch")
    return res


def suboctad_weight(uint32_t u_sub):
    return mat24_suboctad_weight(u_sub)


def suboctad_scalar_prod(uint32_t u_sub1, uint32_t u_sub2):
    return mat24_suboctad_scalar_prod(u_sub1, u_sub2)

###########################################################################
# Represent a cocode element as a subset of a docecad
###########################################################################

def cocode_as_subdodecad(uint32_t c1, uint32_t v1, uint32_t u_single = 24):
    cdef uint32_t res
    res =  mat24_cocode_as_subdodecad(c1, v1, u_single)
    if res & 0xff000000:
        err = "Cocode element is not a subset of the docecad"
        raise ValueError(err)
    return res


###########################################################################
# Parker Loop
###########################################################################


def ploop_theta(uint32_t v1):
    return mat24_ploop_theta(v1)


def ploop_cocycle(uint32_t v1, uint32_t v2):
    return mat24_ploop_cocycle(v1, v2)



def mul_ploop(uint32_t v1, uint32_t v2):
    return mat24_mul_ploop(v1, v2) 


def pow_ploop(uint32_t v1, uint32_t u_exp):
    return mat24_pow_ploop(v1, u_exp) 



def ploop_comm(uint32_t v1, uint32_t v2):
    return mat24_ploop_comm(v1, v2)


def ploop_cap(uint32_t v1, uint32_t v2):
    return mat24_ploop_cap(v1, v2)


def ploop_assoc(uint32_t v1, uint32_t v2, uint32_t v3):
    return mat24_ploop_assoc(v1, v2, v3)


cdef enum:
     ploop_solve_size = 64  

def ploop_solve(a):
    cdef uint32_t pa[ploop_solve_size + 13]
    cdef unsigned int i
    cdef unsigned int res = 0
    cdef unsigned int length = len(a)
    cdef unsigned int start = 0
    cdef unsigned int d
    cdef unsigned int result_length = 0
    while start < length:
        d = min(length - start, ploop_solve_size)
        for i in range(d):
            pa[result_length + i] = a[start + i]
        res = mat24_ploop_solve(pa, result_length + d)
        result_length = res >> 16
        start += d
    if res & 0x1000:
        err = "Cannot correct signs of all Parker loop elements"
        raise ValueError(err)
    return res 

###########################################################################
# Mathieu group M24: conversion of representations
###########################################################################



def perm_complete_heptad(p_io):
    cdef uint8_t p1[24]
    cdef uint8_t *p_p1 = p1
    cdef int i
    for i in range(24): p1[i] = 24 if p_io[i] is None else int(p_io[i]) 
    cdef uint32_t res = mat24_perm_complete_heptad(p_p1)
    if res != 0:
        raise ValueError("Vector is not an umbral heptad")
    for i in range(24): p_io[i]  = p1[i]
    return res



def perm_check(p1):
    cdef uint8_t p1a[24]
    cdef uint8_t *p_p1a = p1a
    cdef int i
    cdef uint32_t acc = 0
    cdef uint32_t entry
    if len(p1) != 24: 
        err = "Permutation in group Mat24 must have length 24"
        raise ValueError(err)
    for i in range(24): 
        entry = int(p1[i])
        p1a[i] = entry
        acc |= entry 
    if acc  & -32 or mat24_perm_check(p_p1a):
        err =   "Permutation is not in group Mat24" 
        raise ValueError(err)


def perm_complete_octad(p_io):
    cdef uint8_t p1[8]
    cdef uint8_t *p_p1 = p1
    cdef int i
    p1[5] = 24
    for i in range(5): p1[i] = p_io[i] 
    if p_io[5] is not None: p1[5] = p_io[5]
    cdef uint32_t res = mat24_perm_complete_octad(p_p1)
    if res:
         err = "Cannot complete a vector to an octad"
         raise ValueError(err)      
    for i in range(8): p_io[i]  = p1[i]


def perm_from_heptads(h1, h2):
    cdef uint8_t h1a[7]
    cdef uint8_t *p_h1a = h1a
    cdef uint8_t h2a[7]
    cdef uint8_t *p_h2a = h2a 
    cdef uint8_t p1[24]
    cdef uint8_t *p_p1 = p1
    cdef int i
    for i in range(7):
        h1a[i] = int(h1[i]) 
        h2a[i] = int(h2[i]) 
    cdef uint32_t res = mat24_perm_from_heptads(p_h1a, p_h2a, p_p1)
    if res:
        err = "Cannot construct permutation in Mat24 from heptads"
        raise ValueError(err)
    return [p1[i] for i in range(24)] 


def perm_from_map(h1, h2):
    cdef uint8_t h1a[24]
    cdef uint8_t *p_h1a = h1a
    cdef uint8_t h2a[24]
    cdef uint8_t *p_h2a = h2a 
    cdef uint8_t p1[24]
    cdef uint8_t *p_p1 = p1
    cdef uint32_t length = len(h1)
    if len(h2) != length:
        err = "Arrays in function perm_from_map() have different lengths"
        raise ValueError(err)
    if length > 24:
        err = "Array in function perm_from_map() is too long"
        raise ValueError(err)
    for i in range(length):
        h1a[i] = int(h1[i]) 
        h2a[i] = int(h2[i]) 
    cdef uint32_t res = mat24_perm_from_map(p_h1a, p_h2a, length, p_p1)
    if res < 0:
        err = "Illegal entry in array in function perm_from_map()"
        raise ValueError(err)
    return res, [p1[i] for i in range(24)] 
        


def m24num_to_perm(uint32_t u_m24):
    cdef uint8_t p1[24]
    cdef uint8_t *p_p1 = p1
    cdef int i
    if mat24_m24num_to_perm(u_m24, p_p1):
        err = "Illegal number for a permutation in Mat24"
        raise ValueError, err
    return [p1[i] for i in range(24)]




def perm_to_m24num(p1):
    cdef uint8_t p1a[24]
    cdef uint8_t *p_p1a = p1a
    cdef int i
    for i in range(24): p1a[i] = int(p1[i]) 
    return  mat24_perm_to_m24num(p_p1a)


def perm_to_matrix(p1):
    cdef uint8_t p1a[24]
    cdef uint8_t *p_p1a = p1a
    cdef uint32_t m1[12]
    cdef uint32_t *p_m1 = m1
    cdef int i
    for i in range(24): p1a[i] = int(p1[i])
    mat24_perm_to_matrix(p_p1a, p_m1)
    return [m1[i] for i in range(12)]


def matrix_to_perm(m1):
    cdef uint32_t m1a[12]
    cdef uint32_t *p_m1a = m1a
    cdef uint8_t p1[24]
    cdef uint8_t *p_p1 = p1
    cdef int i
    for i in range(12): m1a[i] = int(m1[i])
    mat24_matrix_to_perm(p_m1a, p_p1)
    return [p1[i] for i in range(24)]

def matrix_from_mod_omega(m1):
    cdef uint32_t m1a[12]
    cdef uint32_t *p_m1a = m1a
    cdef int i
    for i in range(12): m1a[i] = int(m1[i])
    mat24_matrix_from_mod_omega(m1a)
    for i in range(12): m1[i] = int(m1a[i])


###########################################################################
# Mathieu group M24: Mapping a dodecad
###########################################################################



def perm_from_dodecads(d1, d2):
    cdef uint8_t d1a[9]
    cdef uint8_t *p_d1a = d1a
    cdef uint8_t d2a[9]
    cdef uint8_t *p_d2a = d2a 
    cdef uint8_t p1[24]
    cdef uint8_t *p_p1 = p1
    cdef int i
    for i in range(9):
        d1a[i] = int(d1[i]) 
        d2a[i] = int(d2[i]) 
    cdef uint32_t res = mat24_perm_from_dodecads(p_d1a, p_d2a, p_p1)
    if res:
        err = "Cannot construct permutation in Mat24 from dodecads"
        raise ValueError(err)
    return [p1[i] for i in range(24)] 




###########################################################################
# Mathieu group M24: operation of group elements
###########################################################################


def op_vect_perm(uint32_t v1, p1):
    cdef uint8_t p1a[24]
    cdef uint8_t *p_p1a = p1a
    cdef int i
    for i in range(24): p1a[i] = int(p1[i])
    return mat24_op_vect_perm(v1, p_p1a)



def op_gcode_matrix(uint32_t v1, m1):
    cdef uint32_t m1a[12]
    cdef uint32_t *p_m1a = m1a
    cdef int i
    for i in range(12): m1a[i] = int(m1[i])
    return mat24_op_gcode_matrix(v1,  p_m1a)


def op_gcode_perm(uint32_t v1, p1):
    cdef uint8_t p1a[24]
    cdef uint8_t *p_p1a = p1a
    cdef int i
    for i in range(24): p1a[i] = int(p1[i])
    return mat24_op_gcode_perm(v1, p_p1a)




def op_cocode_perm(uint32_t c1, p1):
    cdef uint8_t p1a[24]
    cdef uint8_t *p_p1a = p1a
    cdef int i
    for i in range(24): p1a[i] = int(p1[i])
    return mat24_op_cocode_perm(c1, p_p1a)



def mul_perm(p1, p2):
    cdef uint8_t pp1[24]
    cdef uint8_t *p_pp1 = pp1
    cdef uint8_t pp2[24]
    cdef uint8_t *p_pp2 = pp2
    cdef uint8_t pp3[24]
    cdef uint8_t *p_pp3 = pp3
    cdef int i
    for i in range(24): 
        pp1[i] = int(p1[i])
        pp2[i] = int(p2[i])
    mat24_mul_perm(p_pp1, p_pp2, p_pp3)
    return [pp3[i] for i in range(24)]


def inv_perm(p1):
    cdef uint8_t pp1[24]
    cdef uint8_t *p_pp1 = pp1
    cdef uint8_t pp2[24]
    cdef uint8_t *p_pp2 = pp2
    cdef int i
    for i in range(24): 
        pp1[i] = int(p1[i])
    mat24_inv_perm(p_pp1, p_pp2)
    return [pp2[i] for i in range(24)]


###########################################################################
# Automorphisms of the Parker Loop
###########################################################################
 


def autpl_set_qform(m_io):
    cdef uint32_t m_ioa[12]
    cdef uint32_t *p_m_ioa = m_ioa
    cdef int i
    for i in range(12): m_ioa[i] = int(m_io[i])
    mat24_autpl_set_qform(p_m_ioa)
    for i in range(12): m_io[i] = int(m_ioa[i])
    return m_io



def perm_to_autpl(c1, p1):
    cdef uint8_t p1a[24]
    cdef uint8_t *p_p1a = p1a
    cdef uint32_t m1[12]
    cdef uint32_t *p_m1 = m1
    cdef int i
    for i in range(24): p1a[i] = int(p1[i])
    mat24_perm_to_autpl(int(c1), p_p1a, p_m1)
    return [m1[i] for i in range(12)]


def cocode_to_autpl(c1):
    cdef uint32_t m1[12]
    cdef uint32_t *p_m1 = m1
    mat24_cocode_to_autpl(int(c1), p_m1)
    return [m1[i] for i in range(12)]



def autpl_to_perm(m1):
    cdef uint32_t m1a[12]
    cdef uint32_t *p_m1a = m1a
    cdef uint8_t p1[24]
    cdef uint8_t *p_p1 = p1
    cdef int i
    for i in range(12): m1a[i] = int(m1[i])
    mat24_autpl_to_perm(p_m1a, p_p1)
    return [p1[i] for i in range(24)]


def autpl_to_cocode(m1):
    cdef uint32_t m1a[12]
    cdef uint32_t *p_m1a = m1a
    cdef int i
    for i in range(12): m1a[i] = int(m1[i])
    return mat24_autpl_to_cocode(p_m1a)



def op_ploop_autpl(v1, m1):
    cdef uint32_t m1a[12]
    cdef uint32_t *p_m1a = m1a
    cdef int i
    for i in range(12): m1a[i] = int(m1[i])
    return mat24_op_ploop_autpl(v1, p_m1a)


def mul_autpl(m1, m2):
    cdef uint32_t m1a[12]
    cdef uint32_t *p_m1a = m1a
    cdef uint32_t m2a[12]
    cdef uint32_t *p_m2a = m2a
    cdef uint32_t m3a[12]
    cdef uint32_t *p_m3a = m3a
    cdef int i
    for i in range(12): 
        m1a[i] = int(m1[i])
        m2a[i] = int(m2[i])
    mat24_mul_autpl(m1a, m2a, m3a)
    return [m3a[i] for i in range(12)]


def inv_autpl(m1):
    cdef uint32_t m1a[12]
    cdef uint32_t *p_m1a = m1a
    cdef uint32_t m2a[12]
    cdef uint32_t *p_m2a = m2a
    cdef int i
    for i in range(12): 
        m1a[i] = int(m1[i])
    mat24_inv_autpl(m1a, m2a)
    return [m2a[i] for i in range(12)]



def perm_to_iautpl(c1, p1):
    cdef uint8_t p1a[24]
    cdef uint8_t *p_p1a = p1a
    cdef uint32_t m1[12]
    cdef uint32_t *p_m1 = m1
    cdef uint8_t pi[24]
    cdef uint8_t *p_pi = pi
    cdef int i
    for i in range(24): p1a[i] = int(p1[i])
    mat24_perm_to_iautpl(int(c1), p_p1a, p_pi, p_m1)
    return [pi[i] for i in range(24)], [m1[i] for i in range(12)]


###########################################################################
# Auxiliary functions for the Monster group
#
# These functions are not available in class mmgroup.dev.mat24_ref.Mat24
###########################################################################


def perm_to_net(p1):
    cdef uint8_t p[24]
    cdef uint8_t *p_p = p
    cdef uint32_t res[9]
    cdef uint32_t *p_res =res
    cdef int i
    for i in range(24):
        p[i] = p1[i]
    mat24_perm_to_net(p_p, p_res)
    return [res[i] for i in range(9)]





def op_all_autpl(m1):
    cdef uint32_t pm[12]
    cdef uint32_t *p_pm = pm
    cdef uint16_t res[2048]
    cdef uint16_t *p_res = res
    cdef int i
    for i in range(12):
        pm[i] = m1[i]
    mat24_op_all_autpl(p_pm, p_res)
    return [res[i] for i in range(2048)]



def op_all_cocode(c1):
    cdef uint8_t res[2048]
    cdef uint8_t *p_res = res
    cdef uint32_t c1i = c1
    cdef int i
    mat24_op_all_cocode(c1i, p_res)
    return [res[i] for i in range(2048)]



########################################################################
########################################################################
# The following stuff is experimental and not documented officially 
########################################################################
########################################################################



cpdef uint32_t Mat24Sub_vtype(uint32_t x):
    """returns type of bit vector in GF(2)**24. 

    This is either (w,d,y), 0 <= d <= 3,  or (w,4,t)

    Here w is the bit weight of vector v. In the first case, there
    is a unique syndrome s(x) of x of weight d, 0 <= d <= 3, and
    the weight of x & s(x) is equal to y in case w <= 12 and to
    d - y in case w >= 12.
    In the second case the syndrome may be any of a tetrad of 
    six vectors s_i of length 4 with disjoint support. In case 
    w = 12, the value  t is the number of vectors s_i completely 
    contained in x. For other values of w the value t is zero.
  
    It is well known that all vectors of the same type are
    in one orbit of the automorphism group of the Golay code.

    The tuple is returned in the form  (w << 5 ) + (d << 2) + y .
    """
    cdef uint32_t w, s, d, y,  rest
    w = mat24_bw24(x)
    s = mat24_syndrome(x,0)
    d = mat24_bw24(s)
    if d < 4: 
        y =  mat24_bw24(x & s) 
        if w >= 12: y = d - y
    elif w != 12: 
        y = 0
    else: 
        y, rest = 0 + (s == x & s),  x & ~s & 0xffffff 
        while rest:
            s =  mat24_syndrome(x, mat24_lsbit24(rest))
            y +=  s == x & s  
            rest &= ~s
    return ( (w << 5 ) + (d << 2) + y )



def Mat24Sub_count_vtypes():
     cdef uint32_t a[8000]
     cdef uint32_t n
     for n in range(8000):
         a[n] = 0 
     for n in range(0x1000000):
         a[Mat24Sub_vtype(n)] += 1     
     d =  {}
     for n in range(8000):
        if a[n] > 0:
            d[ (n>>5, (n>>2) & 7, n & 3) ] = a[n]
     return d

     
       






from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from random import randint 
from functools import reduce

import numpy as np
import pytest

from mmgroup import XLeech2 
from mmgroup import mat24
from mmgroup.generators import gen_leech2_subtype
from mmgroup.generators import gen_leech2_op_atom
from mmgroup.generators import gen_leech2_op_word
from mmgroup.generators import gen_leech2_reduce_n
from mmgroup.generators import gen_leech2_reduce_n_rep

from mmgroup.dev.generators.gen_leech_reduce_n import MAP_VECTOR
from mmgroup.dev.generators.gen_leech_reduce_n import TABLE_OCTAD
from mmgroup.dev.generators.gen_leech_reduce_n import TABLE_DODECAD
from mmgroup.dev.generators.gen_leech_reduce_n import STD_DODECAD
from mmgroup.dev.generators.gen_leech_reduce_n import CPL_DODECAD

from mmgroup.tests.test_gen_xi.test_leech2_type import rand_n_elem
from mmgroup.tests.test_gen_xi.test_reduce import apply_perm
from mmgroup.tests.test_gen_xi.test_reduce import find_octad_permutation


#####################################################################
# Auxiliary functions
#####################################################################

OCTAD_PLUS = [0,1,2,3,4, 8,9]


def find_octad_permutation_odd(v, result, verbose = 0):
    """ Find a suitable permutation for an octad.

    Similar to function ``find_octad_permutation`` in module
    ``mmgroup.dev.generators.gen_leech_reduce_n``.
    Here ``v, o, c`` are as in that function; but the scalar
    product of ``o`` and ``c`` must be 1. Apart from that
    operation is as in function ``find_octad_permutation``.

    We compute a permutation that maps octad ``o`` to the standard
    octad (0,1,2,3,4,5,6,7). If the cocode part ``c`` of ``v`` is 
    not a suboctad of octad ``o`` then we map (one shortest 
    representative of) ``c`` into the set (0,1,2,3,...7,8). 
    """
    coc = (v ^ mat24.ploop_theta(v >> 12)) & 0xfff
    w = mat24.gcode_weight(v >> 12) 
    vect = mat24.gcode_to_vect((v ^ ((w & 4) << 21)) >> 12)
    src = mat24.vect_to_list(vect, 5)
    if mat24.cocode_weight(coc) == 4:
        sextet = mat24.cocode_to_sextet(coc) 
        for i in range(0, 24, 4):
            syn = (1 << sextet[i]) | (1 << sextet[i+1])
            syn |= (1 << sextet[i+2]) | (1 << sextet[i+3])
            special = syn & vect
            if special & (special - 1): 
                break
    else:
        syn = mat24.cocode_syndrome(coc, 24)
    src.append(mat24.lsbit24(syn & ~vect))  
    return apply_perm(v, src, OCTAD_PLUS, 6, result, verbose)


def op_y_x(v, table, out):
    v = gen_leech2_op_atom(v, out[0])
    c = (v ^ table[0]) & 0x7ff
    y = 0
    for i in range(11):
        y ^=  table[i + 1] & -((c >> i) & 1)
    out.append(0xC0000000 + y)
    v = gen_leech2_op_atom(v, out[1])
    if (((v >> 23) ^ table[12]) & 1):
        y =  table[13]
        v = gen_leech2_op_atom(v, 0xC0000000 + y)
        out[1] ^= y         
    c = (v ^ table[0]) & 0x7ff
    x = table[14] if (v & 0x1000000) else 0
    out.append(0x90000000 + x) #!!!







def map_vector(v):
    return MAP_VECTOR[gen_leech2_subtype(v)] if v else 0


#####################################################################
# Python implementation of function to be tested
#####################################################################


def ref_gen_leech2_reduce_n(v, verbose = 0):
    vtype = gen_leech2_subtype(v);
    subtype = vtype & 0xf;

    out = []

    if subtype & 1:
        coc = (v ^  mat24.ploop_theta(v >> 12)) & 0xfff;
        syn = mat24.cocode_syndrome(coc)
        src = mat24.vect_to_list(syn, mat24.bw24(syn))
        assert len(src)  in [1,3]
        lst = [1,2,3] if subtype == 3 else [0]
        apply_perm(v, src, lst, len(src), out)
        v = gen_leech2_op_atom(v, out[0])
        out.append(0xC0000000 + ((v >> 12) & 0x7ff))
        v = gen_leech2_op_atom(v, out[1])
        out.append(0xB0000000 + ((v >> 13) & 0x800))
    elif subtype == 6:
        gv = (v >> 12) & 0xfff 
        vect = mat24.gcode_to_vect(gv)
        src = mat24.vect_to_list(vect, mat24.bw24(vect))
        assert len(src) == 12
        dest = STD_DODECAD
        if (vtype == 0x36):
            coc = (v ^  mat24.ploop_theta(v >> 12)) & 0xfff;
            w =  mat24.bw24(mat24.cocode_as_subdodecad(coc, gv))
            if w & 2:
               dest = CPL_DODECAD 
        pi = mat24.perm_from_dodecads(dest, src)
        out.append(0xA0000000 + mat24.perm_to_m24num(pi))
        op_y_x(v, TABLE_DODECAD, out)
    elif subtype in [2,4]:
        if vtype == 0x34:
            find_octad_permutation_odd(v, out)
        else:
            find_octad_permutation(v, out)
        op_y_x(v, TABLE_OCTAD, out)
    elif subtype in [0,8]:
        if ((v & 0x7ff) == 0):
            out.append(0xA0000000)
            out.append(0xC0000000)
            x = 0x800 if v & 0x1800000 == 0x1800000 else 0
            out.append(0x90000000 + x)
        else:
            syn = mat24.cocode_syndrome(v & 0x7ff, 0)
            src = mat24.vect_to_list(syn, mat24.bw24(syn))
            j = mat24.bw24(syn) & 2;
            lst, y0  = ([2,3], 0x200) if j else ([0,1,2,3], 0x400)
            apply_perm(v, src, lst, len(lst), out);
            v = gen_leech2_op_atom(v, out[0])
            y = y0 if v & 0x800000 else 0
            out.append(0xC0000000 + y)
            v = gen_leech2_op_atom(v, out[1])
            x = y0 if v & 0x1000000 else 0
            out.append(0xB0000000 + x)
    else:
        raise ValueError("Bad subtype " + hex(vtype))
    assert len(out) == 3
    return vtype, np.array(out, dtype = np.uint32)






#####################################################################
# Test data for testing function gen_leech2_reduce_n
#####################################################################


def reduce_n_testdata(ntests = 100):
    """The function yields test vectors in the Leech lattice mod 2

    These test vectors are given in **Leech lattice encoding**;
    they are used for testing function ``gen_leech2_reduce_n``.
    """
    for c in [0,0x200]:
        for d in range(0, 0x2000000, 0x800000):
             yield c + d
    testdata = [
    ]
    for t in testdata:
         yield t
    for i in range(ntests):
        v = randint(0, 0x7ff)
        o = randint(0, 3)
        yield (o << 23) + v
    for v in MAP_VECTOR.values():
        for i in range(ntests):
            g =  rand_n_elem()
            yield g.xsp_conjugate(v)
    for type_ in (2,3,4):      
        for i in range(ntests):
            yield XLeech2('r', type_).ord

#####################################################################
#####################################################################
# Testing function gen_leech2_reduce_n
#####################################################################
#####################################################################




#####################################################################
# Test transversal in function gen_leech2_reduce_n
#####################################################################



@pytest.mark.gen_xi
def test_std_subtypes():
    for subtype, v in [(0,0)] + list(MAP_VECTOR.items()):
        subtype1 =  gen_leech2_subtype(v)
        if subtype1 != subtype:
            print("Subtype obtained: %s, expected: %s" % 
                (hex(subtype1), hex(subtype)))
            print("gcode", gcode)
            print("cocode", cocode)
            err = "Error in transversal of N_x0"
            raise ValueError(err)
    for subtype, v in list(MAP_VECTOR.items()):
        v_c = gen_leech2_reduce_n_rep(subtype)
        assert v == v_c


#####################################################################
# Test function gen_leech2_reduce_n
#####################################################################

@pytest.mark.gen_xi
def test_reduce_n(ntests = 100, verbose = 0):
    """Test function ``gen_leech2_reduce_n`` """
    for n, v in enumerate(reduce_n_testdata(ntests)):
        ref_subtype = gen_leech2_subtype(v)
        if verbose:
            print(" \nTest %d, v = %s, subtype  = %s" % 
                (n+1, hex(v), hex(ref_subtype))
            )
        subtype, op = ref_gen_leech2_reduce_n(v)
        ok = True 
        err = "Unknown error"

        ok1 =  ref_subtype == subtype 
        if ok and not ok1:
            err  = "Function gen_leech2_reduce_n returns wrong subtype"
        ok &= ok1

        v_mapped = gen_leech2_op_word(v, op, len(op))
        ok2 = v_mapped == map_vector(v)
        if ok and not ok2:
            err  = "Wrong mapping in function gen_leech2_reduce_n"
        ok &= ok2

        ok3 = len(op) == 3
        tags = [x >> 28 for x in op]
        ok3 &= tags[:2] == [0xa, 0xc] and tags[2] in [9, 0xb]
        if ok and not ok3:
            err  = "Error in computed operation on vector"
        ok &= ok3

        if  (op[2] & 0xf0000000) == 0xB and v != 0x18000000:
            ok4 = op[2] & 0x800 == 0
        else:
            ok4 = True
        if ok and not ok4:
            err  = "Computed operation on vector must be even"
        ok &= ok4

        op_c = np.zeros(3, dtype = np.uint32)
        subtype_c = gen_leech2_reduce_n(v, op_c) 
        #op_c[1] &= ~2 # Produce an error for verifying the test
        ok_C_function = subtype_c == subtype
        ok_C_function &= (op_c == op).all()
        if ok and not ok_C_function:
            err  = "Error in C function"
        ok &= ok_C_function
    

        if verbose or not ok:
            if not ok:
                print(" \nTest %d, v = %s, subtype  = %s" % 
                    (n+1, hex(v), hex(ref_subtype))
                )
            print("transformation", op)
            print("map expected:", hex(map_vector(v)), ", obtained",
                hex(v_mapped) )
            if not ok_C_function:
                 print("Subtype from C function is %s, expected %s" %
                     (hex(subtype_c), hex(subtype))
                 )
                 print("Operation from C function:", op_c)
                 print("Expected:                 ", op)
            if not ok:
                raise ValueError(err)










from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import pytest

from random import randint, shuffle, sample

from mmgroup import mat24
from mmgroup import GCode, Cocode, PLoop, PLoopZ
from mmgroup import Octad, SubOctad, GcVector 
from mmgroup import AutPL
from mmgroup import Parity


#####################################################################
# Test class Parity
#####################################################################


@pytest.mark.ploop
def test_parity():
    odd, even = Parity(3), Parity(-4)
    assert odd != even
    assert Parity(-311) == odd
    assert int(odd) == 1
    assert int(even) == 0
    assert (-1)**odd == -1
    assert (-1)**even == 1
    assert 1**odd == 1**even == 1
    assert odd + 1 == odd - 1 == 1 + odd == odd + odd == 7 - odd == even
    assert odd * odd == -odd == odd
    assert odd * even == even * odd == 2 * odd == even * 3 == even
    print("\nPossible parities are:", odd,",", even)
    with pytest.raises(ValueError):
        2**even
    with pytest.raises(TypeError):
        even & even
    cc1, cc0 = Cocode(0x823),  Cocode(0x7ef)
    assert cc1 + even == even + cc1 == cc0 + odd == odd + cc0 == odd
    assert cc1 + odd == odd + cc1 == cc0 + even == even + cc0 == even
    assert odd * cc1 == cc1 * odd == cc1 
    ccn = Cocode(0)
    assert even * cc1 == cc1 * even == even * cc0 == cc0 * even == ccn
    pl = PLoop(0x1234)
    gc = GCode(pl)
    assert even * pl == even * gc == GCode(0)
    assert odd * pl == odd * gc == gc
    aut =  AutPL('e', 'r')
    assert odd * aut == odd
    assert even * aut == even

#####################################################################
# Test Parker loop
#####################################################################


def ploop_creation_testvectors(n_cases = 100):
    i = 0
    while i < n_cases:
        v = randint(0, 0xffffff)
        try:
            syn = mat24.syndrome(v)
            sign = randint(0, 1)
            i += 1
            yield sign, v, syn
        except ValueError:
            pass




@pytest.mark.ploop
def test_create_Parker_loop():
    print("\nTesting creation of Parker loop elements")
    for sign, v, syn in ploop_creation_testvectors():
        # Test power map, inveriosn, sign, theta, and conversion to GCode
        vlist = [i for i in range(24) if (v >> i) & 1]
        shuffle(vlist)
        p1 = (-1)**sign * PLoop(vlist)
        ploop = mat24.vect_to_gcode(v ^ syn) + (sign << 12)
        p2 = PLoop(ploop)
        assert p1 == p2
        g1 = GCode(ploop)
        p3 = (-1)**sign * PLoop(g1)
        assert p1 == p3
        assert p1 == PLoop(p3)
        assert abs(p1) == PLoop(GCode(vlist))
        assert abs(p1) == PLoop(GcVector(vlist))
    print( "Parker Loop creation test passed" )



@pytest.mark.ploop
def test_Parker_loop():
    print("\nTesting operation on Parker loop")
    for i in range(200):
        # Test power map, inveriosn, sign, theta, and conversion to GCode
        n1 = randint(0,0x1fff)
        p1 = PLoop(n1)
        assert p1.ord == n1 == p1.gcode + 0x1000 * (1 - p1.sign) / 2
        if (i < 2):
            print("Testing Parker loop element", p1, ", GCode = ", GCode(p1))
        assert len(p1) == mat24.gcode_weight(n1) << 2
        assert p1.theta() == Cocode(mat24.ploop_theta(n1))
        assert p1/4 == p1.theta(p1) == Parity(mat24.ploop_cocycle(n1,n1))
        assert p1**2 == PLoopZ(p1/4) == (-PLoopZ())**(p1/4)
        assert p1**(-5) == PLoopZ(p1/4) * p1 == (-1)**(p1/4) * p1 == 1/p1
        assert (1/p1).ord ^ n1 == (mat24.gcode_weight(n1) & 1) << 12
        assert -1 / p1 == -p1**3 == -(1/p1)
        assert p1 * (-1/p1) == PLoopZ(1) == -PLoopZ()
        assert abs(p1).ord == GCode(p1).ord == p1.ord & 0xfff
        assert p1 != GCode(p1)
        assert +p1 == 1 * p1 == p1 * 1 == p1
        assert p1 != -p1 and p1 != ~p1
        assert (-p1).ord == p1.ord ^ 0x1000
        assert (~p1).ord == p1.ord ^ 0x800
        s, o, p1_pos = p1.split()
        assert s == (p1.ord >> 12) & 1
        assert o == (p1.ord >> 11) & 1
        assert p1.sign == (-1)**s
        assert p1_pos.ord == p1.ord & 0x7ff
        assert PLoopZ(s, o) * p1_pos == p1
        assert PLoopZ(0, o) * p1_pos == abs(p1) 
        assert PLoopZ(s+1, o) * p1_pos == -p1 
        assert -p1 == -1 * p1 == p1 * -1 == p1 / -1 == -1 / p1**-5
        assert PLoopZ(s, 1+o) * p1_pos == ~p1
        assert PLoopZ(1+s, 1+o) * p1_pos == ~-p1 == -~p1 == ~p1 / -1
        assert 2*p1 == GCode(0) ==  -2 *  GCode(p1)
        assert -13*p1 == GCode(p1) == 7 * GCode(p1)
        if len(p1) & 7 == 0:
            assert p1**Parity(1) == p1
            assert p1**Parity(0) == PLoop(0)
        else:
            with pytest.raises(ValueError):
                p1**Parity(randint(0,1))
        assert p1.bit_list == mat24.gcode_to_bit_list(p1.value & 0xfff)
        assert p1.bit_list == GcVector(p1).bit_list
        assert p1.bit_list == GCode(p1).bit_list
        assert PLoop(GcVector(p1)+0) == PLoop(GCode(p1)+0) == abs(p1)
        assert p1 + 0 == 0 + p1 == GCode(p1) + 0 == 0 + GCode(p1)
        assert Cocode(GcVector(p1)) == Cocode(0)
        assert p1/2 == Parity(0)

        # Test Parker loop multiplication and commutator
        n2 = randint(0,0x1fff)
        p2 = PLoop(n2)
        coc = Cocode(mat24.ploop_cap(n1, n2))
        if (i < 1):
            print("Intersection with", p2, "is", coc)
        p2inv = p2**-1
        assert p1 * p2 == PLoop(mat24.mul_ploop(p1.ord, p2.ord))
        assert p1 / p2 == p1 * p2**(-1)
        assert p1 + p2 == p1 - p2 == GCode(p1 * p2) == GCode(n1 ^ n2)
        assert (p1 * p2) / (p2 * p1) == PLoopZ((p1 & p2)/2)
        assert p1 & p2 == coc
        assert p1.theta() == Cocode(mat24.ploop_theta(p1.ord))
        assert p1.theta(p2) == Parity(mat24.ploop_cocycle(p1.ord, p2.ord))
        assert (p1 & p2)/2 == p1.theta(p2) + p2.theta(p1) 
        assert p1 & p2 == p1.theta() + p2.theta() + (p1 + p2).theta()
        assert int((p1 & p2)/2) == mat24.ploop_comm(p1.ord, p2.ord) 
        assert GcVector(p1 & p2) ==  GcVector(p1) &  GcVector(p2)      
        assert ~GcVector(p1 & p2) ==  ~GcVector(p1) |  ~GcVector(p2) 
        assert Cocode(GcVector(p1 & p2)) == p1 & p2       

        # Test associator
        n3 = randint(0,0x1fff)
        p3 = PLoop(n3)
        assert p1 * p2 * p3 / (p1 * (p2 * p3)) == PLoopZ(p1 & p2 & p3)
        assert int(p1 & p2 & p3) == mat24.ploop_assoc(
                p1.ord, p2.ord, p3.ord)
        i = randint(-1000, 1000)
        par = Parity(i)
        s3 = ((p3 & p1) & p2) + par
        assert s3 ==  ((p3 & p1) & p2) + par 
        assert s3 ==  i + (p1 & (p2 & p3))
        assert s3  == par + (p1 & (p2 & p3))
  
        # Test some operations leading to a TypeError
        with pytest.raises(TypeError):
            p1 & p2 & p3 & p1
        with pytest.raises(TypeError):
            coc & coc
        with pytest.raises(TypeError):
            GCode(p1) * GCode(p2)
        with pytest.raises(TypeError):
            1 / GCode(p2)
        with pytest.raises(ValueError):
           coc / 4
        with pytest.raises(TypeError):
           p1 * coc 
        with pytest.raises(TypeError):
           coc * p1
        types = [GcVector, GCode, Cocode, PLoop]
        for type_ in types:
            with pytest.raises(TypeError):
                int(type_(0))

    print( "Parker Loop test passed" )




@pytest.mark.ploop
def test_cocode():
    print("")
    for i in range(200):
        # Test power map, inveriosn, sign, theta, and conversion to GCode
        n1 = randint(0,0x1fff)
        p1 = PLoop(n1)
        ccvector = randint(0, 0xffffff)
        coc = mat24.vect_to_cocode(ccvector)
        cclist =  [i for i in range(24) if (ccvector >> i) & 1]
        cc1 = Cocode(cclist)
        cc2 = Cocode(coc)
        if i < 1:
            print("\nTesting", GcVector(ccvector), ", cocode =", cc1)
        assert cc1 == cc2
        u = Parity(mat24.scalar_prod(p1.value, cc1.value))
        assert p1 & cc1 == u == cc1 & p1 == u * 1 == u + 0
        par = Parity(randint(0,1))
        assert cc1 + par == par + cc1 == cc1.value//0x800 + par
        assert cc1 % 2 == Parity(cc1)
        assert len(cc1) == mat24.cocode_weight(cc1.value)
        if len(cc1) < 4:
            syndrome = mat24.cocode_syndrome(cc1.value)
            assert cc1.syndrome().value ==  syndrome
            syn_from_list = sum(
                1 << i for i in GcVector(ccvector).syndrome_list())
            assert syn_from_list == syndrome
        i = randint(0,23)
        assert cc1.syndrome(i).value ==  mat24.cocode_syndrome(cc1.value, i)
        syndrome_list = cc1.syndrome(i).bit_list
        assert len(cc1) == len(syndrome_list)
        assert syndrome_list == mat24.cocode_to_bit_list(cc1.value, i)
        
        
@pytest.mark.ploop
def test_octads():
    print("")
    OMEGA = ~PLoop(0)
    for i in range(200):
        no = randint(0, 758)
        o = Octad(no)
        assert o == PLoop(mat24.octad_to_gcode(no))
        assert o == Octad(sample(o.bit_list, 5))
        assert o.octad == no
        o_signbit, o_cpl = randint(0,1), randint(0,1)
        signed_o = PLoopZ(o_signbit, o_cpl) * o
        assert signed_o.octad == no
        assert signed_o.sign == (-1)**o_signbit
        assert o.gcode == mat24.octad_to_gcode(no)
        assert signed_o.gcode ==  (o * PLoopZ(0, o_cpl)).gcode
        assert signed_o.split_octad() == (o_signbit, o_cpl, o)
        nsub =  randint(0, 63)
        sub = (-1)**o_signbit * SubOctad(no, nsub)
        sub1 = (-1)**o_signbit * SubOctad(o, nsub)
        sub_weight = mat24.suboctad_weight(nsub)
        assert sub == sub1
        assert o == (-1)**o_signbit * Octad(sub) * OMEGA**sub_weight 
        assert sub.octad_number() == no
        ploop, cocode = sub.isplit()
        sign, gcode = (-1)** (ploop >> 12), ploop & 0xfff
        assert gcode == o.gcode ^ 0x800 * sub_weight  
        #assert sub.suboctad == nsub 
        assert sign == signed_o.sign == (-1)**o_signbit
        coc = mat24.suboctad_to_cocode(nsub, o.gcode)
        assert cocode == coc 
        assert Cocode(sub) == Cocode(coc)
        assert sub == SubOctad(sub.sign * o, Cocode(coc))
        assert sub ==  sub.sign  * SubOctad(no, Cocode(coc))

        o2 =  Octad(randint(0, 758))
        sub2 = SubOctad(o, o2)
        assert Cocode(sub2) == Cocode(o & o2)
        assert len(Cocode(sub2))//2 & 1 == int((o & o2)/2)

        t_sign, t_tag, t_i0, t_i1 = sub.vector_tuple()
        assert t_tag == 'T'
        assert t_sign == sign, (hex(sub.value), t_sign, o_signbit)
        assert t_i0 == no
        assert t_i1 == nsub



